""" CSAPI exposes update functionalities to the Configuration.

    Most of these functions can only be done by administrators
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.ConfigurationClient import ConfigurationClient
from DIRAC.Core.Utilities import List, Time
from DIRAC.Core.Security.X509Chain import X509Chain  # pylint: disable=import-error
from DIRAC.Core.Security import Locations
from DIRAC.ConfigurationSystem.private.Modificator import Modificator
from DIRAC.ConfigurationSystem.Client.Helpers import CSGlobals
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getSites, getCESiteMapping
from DIRAC.ConfigurationSystem.Client.Helpers.Path import cfgPath

__RCSID__ = "$Id$"


class CSAPI(object):
    """CSAPI objects need an initialization phase"""

    def __init__(self):
        """
        Initialization function
        """
        self.csModified = False
        self.__baseSecurity = "/Registry"
        self.__baseResources = "/Resources"

        self.__userDN = ""
        self.__userGroup = ""
        self.__rpcClient = None
        self.__csMod = None

        self.__initialized = S_ERROR("Not initialized")
        self.initialize()
        if not self.__initialized["OK"]:
            gLogger.error(self.__initialized)

    def __getProxyID(self):
        """Get proxy identity information

        :return: S_OK()/S_ERROR()
        """
        proxyLocation = Locations.getProxyLocation()
        if not proxyLocation:
            return S_ERROR("No proxy found!")
        chain = X509Chain()
        if not chain.loadProxyFromFile(proxyLocation):
            return S_ERROR("Can't read proxy!", proxyLocation)
        retVal = chain.getIssuerCert()
        if not retVal["OK"]:
            return S_ERROR("Can't parse proxy!", retVal["Message"])
        idCert = retVal["Value"]
        result = idCert.getSubjectDN()
        if result["OK"]:
            self.__userDN = result["Value"]
            result = chain.getDIRACGroup()
        if not result["OK"]:
            return result
        self.__userGroup = result["Value"]
        return S_OK()

    def __getCertificateID(self):
        """Get certificate identity information

        :return: S_OK()/S_ERROR()
        """
        certLocation = Locations.getHostCertificateAndKeyLocation()
        if not certLocation:
            return S_ERROR("No certificate found!")
        chain = X509Chain()
        retVal = chain.loadChainFromFile(certLocation[0])
        if not retVal["OK"]:
            return S_ERROR("Can't parse certificate!", retVal["Message"])
        result = chain.getIssuerCert()
        if result["OK"]:
            idCert = result["Value"]
            result = idCert.getSubjectDN()
        if not result["OK"]:
            return result
        self.__userDN = result["Value"]
        self.__userGroup = "host"
        return S_OK()

    def initialize(self):
        """API initialization

        :return: S_OK()/S_ERROR()
        """
        if self.__initialized["OK"]:
            return self.__initialized
        res = self.__getCertificateID() if gConfig.useServerCertificate() else self.__getProxyID()
        if not res["OK"]:
            gLogger.error(res["Message"])
            self.__initialized = S_ERROR("Cannot locate client credentials")
            return self.__initialized
        retVal = gConfig.getOption("/DIRAC/Configuration/MasterServer")
        if not retVal["OK"]:
            self.__initialized = S_ERROR("Master server is not known. Is everything initialized?")
            return self.__initialized
        self.__rpcClient = ConfigurationClient(url=gConfig.getValue("/DIRAC/Configuration/MasterServer", ""))
        self.__csMod = Modificator(
            self.__rpcClient,
            "%s - %s - %s" % (self.__userGroup, self.__userDN, Time.dateTime().strftime("%Y-%m-%d %H:%M:%S")),
        )
        retVal = self.downloadCSData()
        if not retVal["OK"]:
            self.__initialized = S_ERROR("Can not download the remote cfg. Is everything initialized?")
            return self.__initialized
        self.__initialized = S_OK()
        return self.__initialized

    def downloadCSData(self):
        """Download CS data

        :return: S_OK()/S_ERROR()
        """
        if not self.__csMod:
            return S_ERROR("CSAPI not yet initialized")
        result = self.__csMod.loadFromRemote()
        if not result["OK"]:
            return result
        self.csModified = False
        self.__csMod.updateGConfigurationData()
        return S_OK()

    # Resources-related methods
    #########################################

    def addSite(self, siteName, optionsDict=None):
        """Adds a new site to the CS.
        A site is a container for services, so after calling this function,
        at least addCEtoSite() should be called.

        :param str siteName: FQN of the site (e.g. LCG.CERN.ch)
        :param dict optionsDict: optional dictionary of options
        :returns: S_OK/S_ERROR structure
        """

        if not self.__initialized["OK"]:
            return self.__initialized

        self.__csMod.createSection(cfgPath(self.__baseResources, "Sites"))
        self.__csMod.createSection(cfgPath(self.__baseResources, "Sites", siteName.split(".")[0]))
        self.__csMod.createSection(cfgPath(self.__baseResources, "Sites", siteName.split(".")[0], siteName))
        # add options if requested
        if optionsDict:
            for option, optionValue in optionsDict.items():  # can be an iterator
                self.__csMod.setOptionValue(
                    cfgPath(self.__baseResources, "Sites", siteName.split(".")[0], siteName, option), optionValue
                )
        self.csModified = True
        return S_OK(True)

    def addCEtoSite(self, siteName, ceName, optionsDict=None):
        """Adds a new CE to a site definition in the CS.
        A CE normally has queues, so addQueueToCE should be called after this one.

        :param str siteName: FQN of the site (e.g. LCG.CERN.ch)
        :param str ceName: FQN of the CE (e.g. ce503.cern.ch)
        :param dict optionsDict: optional dictionary of options
        :returns: S_OK/S_ERROR structure
        """
        res = getSites()
        if not res["OK"]:
            return res
        if siteName not in res["Value"]:
            res = self.addSite(siteName)
            if not res["OK"]:
                return res

        # CSAPI.createSection() always returns S_OK even if the section already exists
        self.__csMod.createSection(cfgPath(self.__baseResources, "Sites", siteName.split(".")[0], siteName, "CEs"))
        self.__csMod.createSection(
            cfgPath(self.__baseResources, "Sites", siteName.split(".")[0], siteName, "CEs", ceName)
        )
        # add options if requested
        if optionsDict is not None:
            for option, optionValue in optionsDict.items():  # can be an iterator
                self.__csMod.setOptionValue(
                    cfgPath(self.__baseResources, "Sites", siteName.split(".")[0], siteName, "CEs", ceName, option),
                    optionValue,
                )
        self.csModified = True
        return S_OK(True)

    def addQueueToCE(self, ceName, queueName, optionsDict=None):
        """Adds a new queue to a CE definition in the CS.

        :param str ceName: FQN of the CE (e.g. ce503.cern.ch)
        :param str queueName: name of the queue (e.g. ce503.cern.ch-condor)
        :param dict optionsDict: optional dictionary of options
        :returns: S_OK/S_ERROR structure
        """
        res = getCESiteMapping(ceName)
        if not res["OK"]:
            return res
        if ceName not in res["Value"]:
            return S_ERROR("CE does not exist")
        siteName = res["Value"][ceName]

        # CSAPI.createSection() always returns S_OK even if the section already exists
        self.__csMod.createSection(
            cfgPath(self.__baseResources, "Sites", siteName.split(".")[0], siteName, "CEs", ceName, "Queues", queueName)
        )
        # add options if requested
        if optionsDict is not None:
            for option, optionValue in optionsDict.items():  # can be an iterator
                self.__csMod.setOptionValue(
                    cfgPath(
                        self.__baseResources,
                        "Sites",
                        siteName.split(".")[0],
                        siteName,
                        "CEs",
                        ceName,
                        "Queues",
                        queueName,
                        option,
                    ),
                    optionValue,
                )
        self.csModified = True
        return S_OK(True)

    # Registry-related methods
    #########################################

    def listUsers(self, group=False):
        """List users

        :param str group: group name

        :return: S_OK(list)/S_ERROR()
        """
        if not self.__initialized["OK"]:
            return self.__initialized
        if not group:
            return S_OK(self.__csMod.getSections("%s/Users" % self.__baseSecurity))
        users = self.__csMod.getValue("%s/Groups/%s/Users" % (self.__baseSecurity, group))
        return S_OK(List.fromChar(users) if users else [])

    def listHosts(self):
        """List hosts

        :return: S_OK(list)/S_ERROR()
        """
        if not self.__initialized["OK"]:
            return self.__initialized
        return S_OK(self.__csMod.getSections("%s/Hosts" % self.__baseSecurity))

    def describeUsers(self, users=None):
        """describe users by nickname

        :param list users: list of users' nicknames

        :return: a S_OK(description) of the users in input
        """
        return self.__describeEntity(users)

    def describeHosts(self, hosts=None):
        """describe hosts

        :param list users: list of users' nicknames

        :return: a S_OK(description) of the users in input
        """
        return self.__describeEntity(hosts, True)

    def __describeEntity(self, mask, hosts=False):
        """Describe users or hosts

        :param list mask: list users or hosts
        :param bool hosts: if its hosts

        :return: S_OK(dict)/S_ERROR()
        """
        if not self.__initialized["OK"]:
            return self.__initialized
        csSection = "%s/%s" % (self.__baseSecurity, ("Hosts" if hosts else "Users"))
        entities = self.__csMod.getSections(csSection)
        if mask:
            entities = [entity for entity in entities if entity in (mask or [])]
        entitiesDict = {}
        for entity in entities:
            entitiesDict[entity] = {}
            for option in self.__csMod.getOptions("%s/%s" % (csSection, entity)):
                entitiesDict[entity][option] = self.__csMod.getValue("%s/%s/%s" % (csSection, entity, option))
            if not hosts:
                result = self.describeGroups()
                if not result["OK"]:
                    return result
                groupsDict = result["Value"]
                entitiesDict[entity]["Groups"] = []
                for group in groupsDict:
                    if "Users" in groupsDict[group] and entity in groupsDict[group]["Users"]:
                        entitiesDict[entity]["Groups"].append(group)
                entitiesDict[entity]["Groups"].sort()
        return S_OK(entitiesDict)

    def listGroups(self):
        """
        List all groups
        """
        if not self.__initialized["OK"]:
            return self.__initialized
        return S_OK(self.__csMod.getSections("%s/Groups" % self.__baseSecurity))

    def describeGroups(self, mask=None):
        """
        List all groups that are in the mask (or all if no mask) with their properties
        """
        if mask is None:
            mask = []
        if not self.__initialized["OK"]:
            return self.__initialized
        groups = [
            group
            for group in self.__csMod.getSections("%s/Groups" % self.__baseSecurity)
            if not mask or (mask and group in mask)
        ]
        groupsDict = {}
        for group in groups:
            groupsDict[group] = {}
            for option in self.__csMod.getOptions("%s/Groups/%s" % (self.__baseSecurity, group)):
                groupsDict[group][option] = self.__csMod.getValue(
                    "%s/Groups/%s/%s" % (self.__baseSecurity, group, option)
                )
                if option in ("Users", "Properties"):
                    groupsDict[group][option] = List.fromChar(groupsDict[group][option])
        return S_OK(groupsDict)

    def deleteUsers(self, users):
        """Delete a user/s can receive as a param either a string or a list

        :param list users: users

        :return: S_OK(bool)/S_ERROR()
        """
        if not self.__initialized["OK"]:
            return self.__initialized
        if isinstance(users, six.string_types):
            users = [users]
        result = self.describeUsers(users)
        if not result["OK"]:
            return result
        usersData = result["Value"]
        for username in users:
            if username not in usersData:
                gLogger.warn("User %s does not exist" % username)
                continue
            userGroups = usersData[username]["Groups"]
            for group in userGroups:
                self.__removeUserFromGroup(group, username)
                gLogger.info("Deleted user %s from group %s" % (username, group))
            self.__csMod.removeSection("%s/Users/%s" % (self.__baseSecurity, username))
            gLogger.info("Deleted user %s" % username)
            self.csModified = True
        return S_OK(True)

    def __removeUserFromGroup(self, group, username):
        """
        Remove user from a group
        """
        usersInGroup = self.__csMod.getValue("%s/Groups/%s/Users" % (self.__baseSecurity, group))
        if usersInGroup is not None:
            userList = List.fromChar(usersInGroup, ",")
            userPos = userList.index(username)
            userList.pop(userPos)
            self.__csMod.setOptionValue("%s/Groups/%s/Users" % (self.__baseSecurity, group), ",".join(userList))

    def __addUserToGroup(self, group, username):
        """
        Add user to a group
        """
        usersInGroup = self.__csMod.getValue("%s/Groups/%s/Users" % (self.__baseSecurity, group))
        if usersInGroup is not None:
            userList = List.fromChar(usersInGroup)
            if username not in userList:
                userList.append(username)
                self.__csMod.setOptionValue("%s/Groups/%s/Users" % (self.__baseSecurity, group), ",".join(userList))
            else:
                gLogger.warn("User %s is already in group %s" % (username, group))

    def addUser(self, username, properties):
        """
        Add a user to the cs

        :param str username: username
        :param dict properties: dictionary describing user properties:

          - DN
          - groups
          - <extra params>

        :return: S_OK(bool)/S_ERROR()
        """
        if not self.__initialized["OK"]:
            return self.__initialized
        for prop in ("DN", "Groups"):
            if prop not in properties:
                gLogger.error("Missing property for user", "%s: %s" % (prop, username))
                return S_OK(False)
        result = self.listUsers()
        if not result["OK"]:
            return result
        if username in result["Value"]:
            gLogger.error("User is already registered", username)
            return S_OK(False)
        result = self.listGroups()
        if not result["OK"]:
            return result
        groups = result["Value"]
        for userGroup in properties["Groups"]:
            if userGroup not in groups:
                gLogger.error("User group is not a valid group", "%s %s" % (username, userGroup))
                return S_OK(False)
        self.__csMod.createSection("%s/Users/%s" % (self.__baseSecurity, username))
        for prop in properties:
            if prop == "Groups":
                continue
            self.__csMod.setOptionValue("%s/Users/%s/%s" % (self.__baseSecurity, username, prop), properties[prop])
        for userGroup in properties["Groups"]:
            gLogger.info("Added user %s to group %s" % (username, userGroup))
            self.__addUserToGroup(userGroup, username)
        gLogger.info("Registered user %s" % username)
        self.csModified = True
        return S_OK(True)

    def modifyUser(self, username, properties, createIfNonExistant=False):
        """
        Modify a user

        :param str username: group name
        :param dict properties: dictionary describing user properties:

            - DN
            - Groups
            - <extra params>

        :param bool createIfNonExistant: if true, registers the users if it did not exist
        :return: S_OK, Value = True/False
        """
        if not self.__initialized["OK"]:
            return self.__initialized
        modifiedUser = False
        result = self.describeUsers([username])
        if not result["OK"]:
            return result
        userData = result["Value"]
        if username not in userData:
            if createIfNonExistant:
                gLogger.info("Registering user %s" % username)
                return self.addUser(username, properties)
            gLogger.error("User is not registered: ", repr(username))
            return S_OK(False)
        for prop in properties:
            if prop == "Groups":
                continue
            prevVal = self.__csMod.getValue("%s/Users/%s/%s" % (self.__baseSecurity, username, prop))
            if not prevVal or prevVal != properties[prop]:
                gLogger.info("Setting %s property for user %s to %s" % (prop, username, properties[prop]))
                self.__csMod.setOptionValue("%s/Users/%s/%s" % (self.__baseSecurity, username, prop), properties[prop])
                modifiedUser = True
        if "Groups" in properties:
            result = self.listGroups()
            if not result["OK"]:
                return result
            groups = result["Value"]
            for userGroup in properties["Groups"]:
                if userGroup not in groups:
                    gLogger.error("User group is not a valid group", "%s %s" % (username, userGroup))
                    return S_OK(False)
            groupsToBeDeletedFrom = []
            groupsToBeAddedTo = []
            for prevGroup in userData[username]["Groups"]:
                if prevGroup not in properties["Groups"]:
                    groupsToBeDeletedFrom.append(prevGroup)
                    modifiedUser = True
            for newGroup in properties["Groups"]:
                if newGroup not in userData[username]["Groups"]:
                    groupsToBeAddedTo.append(newGroup)
                    modifiedUser = True
            for group in groupsToBeDeletedFrom:
                self.__removeUserFromGroup(group, username)
                gLogger.info("Removed user %s from group %s" % (username, group))
            for group in groupsToBeAddedTo:
                self.__addUserToGroup(group, username)
                gLogger.info("Added user %s to group %s" % (username, group))
        modified = False
        if modifiedUser:
            modified = True
            gLogger.info("Modified user %s" % username)
            self.csModified = True
        else:
            gLogger.info("Nothing to modify for user %s" % username)
        return S_OK(modified)

    def addGroup(self, groupname, properties):
        """
        Add a group to the cs

        :param str groupname: group name
        :param dict properties: dictionary describing group properties:

            - Users
            - Properties
            - <extra params>

        :return: S_OK, Value = True/False
        """
        if not self.__initialized["OK"]:
            return self.__initialized
        result = self.listGroups()
        if not result["OK"]:
            return result
        if groupname in result["Value"]:
            gLogger.error("Group is already registered", groupname)
            return S_OK(False)
        self.__csMod.createSection("%s/Groups/%s" % (self.__baseSecurity, groupname))
        for prop in properties:
            self.__csMod.setOptionValue("%s/Groups/%s/%s" % (self.__baseSecurity, groupname, prop), properties[prop])
        gLogger.info("Registered group %s" % groupname)
        self.csModified = True
        return S_OK(True)

    def modifyGroup(self, groupname, properties, createIfNonExistant=False):
        """
        Modify a group

        :param str groupname: group name
        :param dict properties: dictionary describing group properties:

            - Users
            - Properties
            - <extra params>

        :param bool createIfNonExistant: if true, creates the group if it did not exist
        :return: S_OK, Value = True/False
        """
        if not self.__initialized["OK"]:
            return self.__initialized
        modifiedGroup = False
        result = self.describeGroups([groupname])
        if not result["OK"]:
            return result
        groupData = result["Value"]
        if groupname not in groupData:
            if createIfNonExistant:
                gLogger.info("Registering group %s" % groupname)
                return self.addGroup(groupname, properties)
            gLogger.error("Group is not registered", groupname)
            return S_OK(False)
        for prop in properties:
            prevVal = self.__csMod.getValue("%s/Groups/%s/%s" % (self.__baseSecurity, groupname, prop))
            if not prevVal or prevVal != properties[prop]:
                gLogger.info("Setting %s property for group %s to %s" % (prop, groupname, properties[prop]))
                self.__csMod.setOptionValue(
                    "%s/Groups/%s/%s" % (self.__baseSecurity, groupname, prop), properties[prop]
                )
                modifiedGroup = True
        if modifiedGroup:
            gLogger.info("Modified group %s" % groupname)
            self.csModified = True
        else:
            gLogger.info("Nothing to modify for group %s" % groupname)
        return S_OK(True)

    def addHost(self, hostname, properties):
        """
        Add a host to the cs

        :param str hostname: host name
        :param dict properties: dictionary describing host properties:

            - DN
            - Properties
            - <extra params>

        :return: S_OK, Value = True/False
        """
        if not self.__initialized["OK"]:
            return self.__initialized
        for prop in ("DN",):
            if prop not in properties:
                gLogger.error("Missing property for host", "%s %s" % (prop, hostname))
                return S_OK(False)
        result = self.listHosts()
        if not result["OK"]:
            return result
        if hostname in result["Value"]:
            gLogger.error("Host is already registered", hostname)
            return S_OK(False)
        self.__csMod.createSection("%s/Hosts/%s" % (self.__baseSecurity, hostname))
        for prop in properties:
            self.__csMod.setOptionValue("%s/Hosts/%s/%s" % (self.__baseSecurity, hostname, prop), properties[prop])
        gLogger.info("Registered host %s" % hostname)
        self.csModified = True
        return S_OK(True)

    def addShifter(self, shifters=None):
        """
        Adds or modify one or more shifters. Also, adds the shifter section in case this is not present.
        Shifter identities are used in several places, mostly for running agents

        :param dict shifters: has to be in the form {'ShifterRole':{'User':'aUserName', 'Group':'aDIRACGroup'}}

        :return: S_OK/S_ERROR
        """

        def getOpsSection():
            """
            Where is the shifters section?
            """
            vo = CSGlobals.getVO()
            setup = CSGlobals.getSetup()

            if vo:
                res = gConfig.getSections("/Operations/%s/%s/Shifter" % (vo, setup))
                if res["OK"]:
                    return S_OK("/Operations/%s/%s/Shifter" % (vo, setup))

                res = gConfig.getSections("/Operations/%s/Defaults/Shifter" % vo)
                if res["OK"]:
                    return S_OK("/Operations/%s/Defaults/Shifter" % vo)

            else:
                res = gConfig.getSections("/Operations/%s/Shifter" % setup)
                if res["OK"]:
                    return S_OK("/Operations/%s/Shifter" % setup)

                res = gConfig.getSections("/Operations/Defaults/Shifter")
                if res["OK"]:
                    return S_OK("/Operations/Defaults/Shifter")

            return S_ERROR("No shifter section")

        if shifters is None:
            shifters = {}
        if not self.__initialized["OK"]:
            return self.__initialized

        # get current shifters
        opsH = Operations()
        currentShifterRoles = opsH.getSections("Shifter")
        if not currentShifterRoles["OK"]:
            # we assume the shifter section is not present
            currentShifterRoles = []
        else:
            currentShifterRoles = currentShifterRoles["Value"]
        currentShiftersDict = {}
        for currentShifterRole in currentShifterRoles:
            currentShifter = opsH.getOptionsDict("Shifter/%s" % currentShifterRole)
            if not currentShifter["OK"]:
                return currentShifter
            currentShifter = currentShifter["Value"]
            currentShiftersDict[currentShifterRole] = currentShifter

        # Removing from shifters what does not need to be changed
        for sRole in list(shifters):  # note the pop below
            if sRole in currentShiftersDict:
                if currentShiftersDict[sRole] == shifters[sRole]:
                    shifters.pop(sRole)

        # get shifters section to modify
        section = getOpsSection()

        # Is this section present?
        if not section["OK"]:
            if section["Message"] == "No shifter section":
                gLogger.warn(section["Message"])
                gLogger.info("Adding shifter section")
                vo = CSGlobals.getVO()
                if vo:
                    section = "/Operations/%s/Defaults/Shifter" % vo
                else:
                    section = "/Operations/Defaults/Shifter"
                res = self.__csMod.createSection(section)
                if not res:
                    gLogger.error("Section %s not created" % section)
                    return S_ERROR("Section %s not created" % section)
            else:
                gLogger.error(section["Message"])
                return section
        else:
            section = section["Value"]

        # add or modify shifters
        for shifter in shifters:
            self.__csMod.removeSection(section + "/" + shifter)
            self.__csMod.createSection(section + "/" + shifter)
            self.__csMod.createSection(section + "/" + shifter + "/" + "User")
            self.__csMod.createSection(section + "/" + shifter + "/" + "Group")
            self.__csMod.setOptionValue(section + "/" + shifter + "/" + "User", shifters[shifter]["User"])
            self.__csMod.setOptionValue(section + "/" + shifter + "/" + "Group", shifters[shifter]["Group"])

        self.csModified = True
        return S_OK(True)

    def modifyHost(self, hostname, properties, createIfNonExistant=False):
        """
        Modify a host

        :param str hostname: hostname name
        :param dict properties: dictionary describing host properties:

            - DN
            - Properties
            - <extra params>

        :param bool createIfNonExistant: if true, creates the host if it did not exist
        :return: S_OK, Value = True/False
        """
        if not self.__initialized["OK"]:
            return self.__initialized
        modifiedHost = False
        result = self.describeHosts([hostname])
        if not result["OK"]:
            return result
        hostData = result["Value"]
        if hostname not in hostData:
            if createIfNonExistant:
                gLogger.info("Registering host %s" % hostname)
                return self.addHost(hostname, properties)
            gLogger.error("Host is not registered", hostname)
            return S_OK(False)
        for prop in properties:
            prevVal = self.__csMod.getValue("%s/Hosts/%s/%s" % (self.__baseSecurity, hostname, prop))
            if not prevVal or prevVal != properties[prop]:
                gLogger.info("Setting %s property for host %s to %s" % (prop, hostname, properties[prop]))
                self.__csMod.setOptionValue("%s/Hosts/%s/%s" % (self.__baseSecurity, hostname, prop), properties[prop])
                modifiedHost = True
        if modifiedHost:
            gLogger.info("Modified host %s" % hostname)
            self.csModified = True
        else:
            gLogger.info("Nothing to modify for host %s" % hostname)
        return S_OK(True)

    def syncUsersWithCFG(self, usersCFG):
        """
        Sync users with the cfg contents. Usernames have to be sections containing
        DN, Groups, and extra properties as parameters
        """
        if not self.__initialized["OK"]:
            return self.__initialized
        done = True
        for user in usersCFG.listSections():
            properties = {}
            propList = usersCFG[user].listOptions()
            for prop in propList:
                if prop == "Groups":
                    properties[prop] = List.fromChar(usersCFG[user][prop])
                else:
                    properties[prop] = usersCFG[user][prop]
            if not self.modifyUser(user, properties, createIfNonExistant=True):
                done = False
        return S_OK(done)

    def sortUsersAndGroups(self):
        self.__csMod.sortAlphabetically("%s/Users" % self.__baseSecurity)
        self.__csMod.sortAlphabetically("%s/Hosts" % self.__baseSecurity)
        for group in self.__csMod.getSections("%s/Groups" % self.__baseSecurity):
            usersOptionPath = "%s/Groups/%s/Users" % (self.__baseSecurity, group)
            users = self.__csMod.getValue(usersOptionPath)
            if users:
                usersList = sorted(List.fromChar(users))
                sortedUsers = ", ".join(usersList)
                if users != sortedUsers:
                    self.__csMod.setOptionValue(usersOptionPath, sortedUsers)

    def checkForUnexistantUsersInGroups(self):
        allUsers = self.__csMod.getSections("%s/Users" % self.__baseSecurity)
        allGroups = self.__csMod.getSections("%s/Groups" % self.__baseSecurity)
        for group in allGroups:
            usersInGroup = self.__csMod.getValue("%s/Groups/%s/Users" % (self.__baseSecurity, group))
            if usersInGroup:
                filteredUsers = []
                usersInGroup = List.fromChar(usersInGroup)
                for user in usersInGroup:
                    if user in allUsers:
                        filteredUsers.append(user)
                self.__csMod.setOptionValue(
                    "%s/Groups/%s/Users" % (self.__baseSecurity, group), ",".join(filteredUsers)
                )

    def commitChanges(self, sortUsers=True):
        if not self.__initialized["OK"]:
            return self.__initialized
        if self.csModified:
            self.checkForUnexistantUsersInGroups()
            if sortUsers:
                self.sortUsersAndGroups()
            retVal = self.__csMod.commit()
            if not retVal["OK"]:
                gLogger.error("Can't commit new configuration data", "%s" % retVal["Message"])
                return retVal
            return self.downloadCSData()
        return S_OK()

    def commit(self):
        """Commit the accumulated changes to the CS server"""
        if not self.__initialized["OK"]:
            return self.__initialized
        if self.csModified:
            retVal = self.__csMod.commit()
            if not retVal["OK"]:
                gLogger.error("Can't commit new configuration data", "%s" % retVal["Message"])
                return retVal
            return self.downloadCSData()
        return S_OK()

    def mergeFromCFG(self, cfg):
        """Merge the internal CFG data with the input"""
        if not self.__initialized["OK"]:
            return self.__initialized
        self.__csMod.mergeFromCFG(cfg)
        self.csModified = True
        return S_OK()

    def modifyValue(self, optionPath, newValue):
        """Modify an existing value at the specified options path."""
        if not self.__initialized["OK"]:
            return self.__initialized
        prevVal = self.__csMod.getValue(optionPath)
        if prevVal is None:
            return S_ERROR("Trying to set %s to %s but option does not exist" % (optionPath, newValue))
        gLogger.verbose("Changing %s from \n%s \nto \n%s" % (optionPath, prevVal, newValue))
        self.__csMod.setOptionValue(optionPath, newValue)
        self.csModified = True
        return S_OK("Modified %s" % optionPath)

    def setOption(self, optionPath, optionValue):
        """Create an option at the specified path."""
        if not self.__initialized["OK"]:
            return self.__initialized
        self.__csMod.setOptionValue(optionPath, optionValue)
        self.csModified = True
        return S_OK("Created new option %s = %s" % (optionPath, optionValue))

    def setOptionComment(self, optionPath, comment):
        """Create an option at the specified path."""
        if not self.__initialized["OK"]:
            return self.__initialized
        self.__csMod.setComment(optionPath, comment)
        self.csModified = True
        return S_OK("Set option comment %s : %s" % (optionPath, comment))

    def delOption(self, optionPath):
        """Delete an option"""
        if not self.__initialized["OK"]:
            return self.__initialized
        if not self.__csMod.removeOption(optionPath):
            return S_ERROR("Couldn't delete option %s" % optionPath)
        self.csModified = True
        return S_OK("Deleted option %s" % optionPath)

    def createSection(self, sectionPath, comment=""):
        """Create a new section"""
        if not self.__initialized["OK"]:
            return self.__initialized
        self.__csMod.createSection(sectionPath)
        self.csModified = True
        if comment:
            self.__csMod.setComment(sectionPath, comment)
        return S_OK()

    def delSection(self, sectionPath):
        """Delete a section"""
        if not self.__initialized["OK"]:
            return self.__initialized
        if not self.__csMod.removeSection(sectionPath):
            return S_ERROR("Could not delete section %s " % sectionPath)
        self.csModified = True
        return S_OK()

    def copySection(self, originalPath, targetPath):
        """Copy a whole section to a new location"""
        if not self.__initialized["OK"]:
            return self.__initialized
        cfg = self.__csMod.getCFG()
        sectionCfg = cfg[originalPath]
        result = self.createSection(targetPath)
        if not result["OK"]:
            return result
        if not self.__csMod.mergeSectionFromCFG(targetPath, sectionCfg):
            return S_ERROR("Could not merge cfg into section %s" % targetPath)
        self.csModified = True
        return S_OK()

    def moveSection(self, originalPath, targetPath):
        """Move a whole section to a new location"""
        result = self.copySection(originalPath, targetPath)
        if not result["OK"]:
            return result
        result = self.delSection(originalPath)
        if not result["OK"]:
            return result
        self.csModified = True
        return S_OK()

    def mergeCFGUnderSection(self, sectionPath, cfg):
        """Merge the given cfg under a certain section"""
        if not self.__initialized["OK"]:
            return self.__initialized
        result = self.createSection(sectionPath)
        if not result["OK"]:
            return result
        if not self.__csMod.mergeSectionFromCFG(sectionPath, cfg):
            return S_ERROR("Could not merge cfg into section %s" % sectionPath)
        self.csModified = True
        return S_OK()

    def mergeWithCFG(self, cfg):
        """Merge the given cfg with the current config"""
        if not self.__initialized["OK"]:
            return self.__initialized
        self.__csMod.mergeFromCFG(cfg)
        self.csModified = True
        return S_OK()

    def getCurrentCFG(self):
        """Get the current CFG as it is"""
        if not self.__initialized["OK"]:
            return self.__initialized
        return S_OK(self.__csMod.getCFG())

    def showDiff(self):
        """Just shows the differences accumulated within the Modificator object"""
        diffData = self.__csMod.showCurrentDiff()
        gLogger.notice("Accumulated diff with master CS")
        for line in diffData:
            if line[0] in ("+", "-"):
                gLogger.notice(line)

    def forceGlobalConfigurationUpdate(self):
        """
        Force global update of configuration on all the registered services

        :return: S_OK/S_ERROR
        """

        return self.__rpcClient.forceGlobalConfigurationUpdate()



"""
  DictCache.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import datetime
import threading

# DIRAC
from DIRAC.Core.Utilities.LockRing import LockRing


class ThreadLocalDict(threading.local):
    """This class is just useful to have a mutable object (in this case, a dict) as a thread local
    Read the _threading_local docstring for more details.

    Its purpose is to have a different cache per thread
    """

    def __init__(self):  # pylint: disable=super-init-not-called
        """c'tor"""
        # Note: it is on purpose that the threading.local constructor is not called
        # Dictionary, local to a thread, that will be used as such
        self.cache = {}


class MockLockRing(object):
    """This mock class is just used to expose the acquire and release method"""

    def doNothing(self, *args, **kwargs):
        """Really does nothing !"""
        pass

    acquire = release = doNothing


class DictCache(object):
    """DictCache is a generic cache implementation.
    The user can decide whether this cache should be shared among the threads or not, but it is always thread safe
    Note that when shared, the access to the cache is protected by a lock, but not necessarily the
    object you are retrieving from it.
    """

    def __init__(self, deleteFunction=False, threadLocal=False):
        """Initialize the dict cache.

        :param deleteFunction: if not False, invoked when deleting a cached object
        :param threadLocal: if False, the cache will be shared among all the threads, otherwise,
                            each thread gets its own cache.
        """

        self.__threadLocal = threadLocal

        # Placeholder either for a LockRing if the cache is shared,
        # or a mock class if not.
        self.__lock = None

        # One of the following two objects is returned
        # by the __cache property, depending on the threadLocal strategy

        # This is the Placeholder for a shared cache
        self.__sharedCache = {}
        # This is the Placeholder for a shared cache
        self.__threadLocalCache = ThreadLocalDict()

        # Function to clean the elements
        self.__deleteFunction = deleteFunction

    @property
    def lock(self):
        """Return the lock.
        In practice, if the cache is shared among threads, it is a LockRing.
        Otherwise, it is just a mock object.
        """

        if not self.__lock:
            if not self.__threadLocal:
                self.__lock = LockRing().getLock(self.__class__.__name__, recursive=True)
            else:
                self.__lock = MockLockRing()

        return self.__lock

    @property
    def __cache(self):
        """Returns either a shared or a thread local cache.
        In any case, the returned object is a dictionary
        """
        if self.__threadLocal:
            return self.__threadLocalCache.cache

        return self.__sharedCache

    def exists(self, cKey, validSeconds=0):
        """Returns True/False if the key exists for the given number of seconds

        :param cKey: identification key of the record
        :param int validSeconds: The amount of seconds the key has to be valid for

        :return: bool
        """
        self.lock.acquire()
        try:
            # Is the key in the cache?
            if cKey in self.__cache:
                expTime = self.__cache[cKey]["expirationTime"]
                # If it's valid return True!
                if expTime > datetime.datetime.now() + datetime.timedelta(seconds=validSeconds):
                    return True
                else:
                    # Delete expired
                    self.delete(cKey)
            return False
        finally:
            self.lock.release()

    def delete(self, cKey):
        """Delete a key from the cache

        :param cKey: identification key of the record
        """
        self.lock.acquire()
        try:
            if cKey not in self.__cache:
                return
            if self.__deleteFunction:
                self.__deleteFunction(self.__cache[cKey]["value"])
            del self.__cache[cKey]
        finally:
            self.lock.release()

    def add(self, cKey, validSeconds, value=None):
        """Add a record to the cache

        :param cKey: identification key of the record
        :param int validSeconds: valid seconds of this record
        :param value: value of the record
        """
        if max(0, validSeconds) == 0:
            return
        self.lock.acquire()
        try:
            vD = {"expirationTime": datetime.datetime.now() + datetime.timedelta(seconds=validSeconds), "value": value}
            self.__cache[cKey] = vD
        finally:
            self.lock.release()

    def get(self, cKey, validSeconds=0):
        """Get a record from the cache

        :param cKey: identification key of the record
        :param int validSeconds: The amount of seconds the key has to be valid for

        :return: None or value of key
        """
        self.lock.acquire()
        try:
            # Is the key in the cache?
            if cKey in self.__cache:
                expTime = self.__cache[cKey]["expirationTime"]
                # If it's valid return True!
                if expTime > datetime.datetime.now() + datetime.timedelta(seconds=validSeconds):
                    return self.__cache[cKey]["value"]
                else:
                    # Delete expired
                    self.delete(cKey)
            return None
        finally:
            self.lock.release()

    def showContentsInString(self):
        """Return a human readable string to represent the contents

        :return: str
        """
        self.lock.acquire()
        try:
            data = []
            for cKey in self.__cache:
                data.append("%s:" % str(cKey))
                data.append("\tExp: %s" % self.__cache[cKey]["expirationTime"])
                if self.__cache[cKey]["value"]:
                    data.append("\tVal: %s" % self.__cache[cKey]["value"])
            return "\n".join(data)
        finally:
            self.lock.release()

    def getKeys(self, validSeconds=0):
        """Get keys for all contents

        :param int validSeconds: valid time in seconds

        :return: list
        """
        self.lock.acquire()
        try:
            keys = []
            limitTime = datetime.datetime.now() + datetime.timedelta(seconds=validSeconds)
            for cKey in self.__cache:
                if self.__cache[cKey]["expirationTime"] > limitTime:
                    keys.append(cKey)
            return keys
        finally:
            self.lock.release()

    def purgeExpired(self, expiredInSeconds=0):
        """Purge all entries that are expired or will be expired in <expiredInSeconds>

        :param int expiredInSeconds: expired time in a seconds
        """
        self.lock.acquire()
        try:
            keys = []
            limitTime = datetime.datetime.now() + datetime.timedelta(seconds=expiredInSeconds)
            for cKey in self.__cache:
                if self.__cache[cKey]["expirationTime"] < limitTime:
                    keys.append(cKey)
            for cKey in keys:
                if self.__deleteFunction:
                    self.__deleteFunction(self.__cache[cKey]["value"])
                del self.__cache[cKey]
        finally:
            self.lock.release()

    def purgeAll(self, useLock=True):
        """Purge all entries
        CAUTION: useLock parameter should ALWAYS be True except when called from __del__

        :param bool useLock: use lock
        """
        if useLock:
            self.lock.acquire()
        try:
            for cKey in list(self.__cache):
                if self.__deleteFunction:
                    self.__deleteFunction(self.__cache[cKey]["value"])
                del self.__cache[cKey]
        finally:
            if useLock:
                self.lock.release()

    def __del__(self):
        """When the DictCache is deleted, all the entries should be purged.
        This is particularly useful when the DictCache manages files
        CAUTION: if you carefully read the python doc, you will see all the
        caveat of __del__. In particular, no guaranty that it is called...
        (https://docs.python.org/2/reference/datamodel.html#object.__del__)
        """
        self.purgeAll(useLock=False)
        del self.__lock
        if self.__threadLocal:
            del self.__threadLocalCache
        else:
            del self.__sharedCache



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# author: lintao

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import readline

readline.set_completer_delims(" \t\n`~!@#$%^&*()=+[{]}\\|;:'\",<>/?")

import os.path


class DirectoryCompletion(object):
    def __init__(self, fs):
        self.fs = fs

    def parse_text_line(self, text, line, cwd):
        """Here, the text and line is from the
        complete_xxx(self, text, line, begidx, endidx).

        text is the word. line is the whole sentence.

        such as

        /home/ihep/work/initrd.lz[]

        [] indicates the CURSOR.

        so, initrd.lz is the text,
        and ``/home/ihep/work/initrd.lz`` is the line.
        """

        result = []

        path = self.generate_absolute(line, cwd)
        dirname = self.get_dirname(path)
        filename = self.get_filename(path, dirname)

        result = list(self.fs.list_dir(dirname))

        text = filename  # + text

        result = [i for i in result if i.startswith(text)]

        return result

    # check absolute path
    def check_absolute(self, path):
        if path.startswith(self.fs.seq):
            return True
        else:
            return False

    # generate absolute path
    def generate_absolute(self, path, cwd):
        if self.check_absolute(path):
            pass
        else:
            path = os.path.join(cwd, path)
        return path

    # get the parent directory or the current directory
    # Using the last char "/" to determine
    def get_dirname(self, path):
        if self.check_absolute(path):
            # if it is the absolute path,
            # return dirname
            if path.endswith(self.fs.seq):
                path = os.path.normpath(path) + self.fs.seq
            else:
                path = os.path.normpath(os.path.dirname(path)) + self.fs.seq
        path = path.replace("//", "/")
        return path

    def get_filename(self, path, dirname):
        if self.check_absolute(path):
            if path.endswith(self.fs.seq):
                path = ""
            else:
                path = os.path.normpath(os.path.basename(path))
        path = path.replace("//", "/")
        return path


if __name__ == "__main__":
    from DIRAC.DataManagementSystem.Client.CmdDirCompletion.AbstractFileSystem import UnixLikeFileSystem

    ulfs = UnixLikeFileSystem()
    dc = DirectoryCompletion(ulfs)

    print(dc.parse_text_line("ls", "/bin/", "/home/ihep"))
    print(dc.parse_text_line(".vim", "", "/home/ihep"))
    print(dc.parse_text_line("", "", "/home/ihep"))



""" PlotBase is a base class for various Graphs plots

    The DIRAC Graphs package is derived from the GraphTool plotting package of the
    CMS/Phedex Project by ... <to be added>
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from DIRAC.Core.Utilities.Graphs.Palette import Palette
from DIRAC.Core.Utilities.Graphs.GraphData import GraphData
from DIRAC.Core.Utilities.Graphs.GraphUtilities import pixelToPoint, evalPrefs
from matplotlib.axes import Axes
from matplotlib.pylab import setp

__RCSID__ = "$Id$"


class PlotBase(object):
    def __init__(self, data=None, axes=None, *aw, **kw):

        self.ax_contain = axes
        self.canvas = None
        self.figure = None
        if self.ax_contain:
            self.figure = self.ax_contain.get_figure()
            self.canvas = self.figure.canvas
            self.dpi = self.ax_contain.figure.get_dpi()
            self.ax_contain.set_axis_off()
        self.prefs = evalPrefs(*aw, **kw)
        self.coords = {}
        self.palette = Palette()
        if isinstance(data, dict):
            self.gdata = GraphData(data)
        elif isinstance(data, object) and data.__class__ == GraphData:
            self.gdata = data

    def dumpPrefs(self):

        for key in self.prefs:
            print(key.rjust(20), ":", str(self.prefs[key]).ljust(40))

    def setAxes(self, axes):

        self.ax_contain = axes
        self.ax_contain.set_axis_off()
        self.figure = self.ax_contain.get_figure()
        self.canvas = self.figure.canvas
        self.dpi = self.ax_contain.figure.get_dpi()

    def draw(self):

        prefs = self.prefs
        dpi = self.ax_contain.figure.get_dpi()

        # Update palette
        palette = prefs.get("colors", {})
        if palette:
            self.palette.addPalette(palette)

        xlabel = prefs.get("xlabel", "")
        ylabel = prefs.get("ylabel", "")
        xticks_flag = prefs.get("xticks", True)
        yticks_flag = prefs.get("yticks", True)

        text_size = prefs["text_size"]
        text_padding = prefs["text_padding"]

        label_text_size = prefs.get("label_text_size", text_size)
        label_text_size_point = pixelToPoint(label_text_size, dpi)
        tick_text_size = prefs.get("tick_text_size", text_size)
        tick_text_size_point = pixelToPoint(tick_text_size, dpi)

        ytick_length = prefs.get("ytick_length", 7 * tick_text_size)

        plot_title = prefs.get("plot_title", "")
        if not plot_title or plot_title == "NoTitle":
            plot_title_size = 0
            plot_title_padding = 0
        else:
            plot_title_size = prefs.get("plot_title_size", text_size)
            plot_title_padding = prefs.get("plot_text_padding", text_padding)
        plot_title_size_point = pixelToPoint(plot_title_size, dpi)

        stats_flag = prefs.get("statistics_line", False)
        stats_line = ""
        stats_line_space = 0.0
        if stats_flag:
            stats_line = self.gdata.getStatString()
            stats_line_size = label_text_size
            stats_line_padding = label_text_size * 2.0
            stats_line_space = stats_line_size + stats_line_padding

        plot_padding = prefs["plot_padding"]
        plot_left_padding = prefs.get("plot_left_padding", plot_padding)
        plot_right_padding = prefs.get("plot_right_padding", 0)
        plot_bottom_padding = prefs.get("plot_bottom_padding", plot_padding)
        plot_top_padding = prefs.get("plot_top_padding", 0)
        frame_flag = prefs["frame"]

        # Create plot axes, and set properties
        left, bottom, width, height = self.ax_contain.get_window_extent().bounds
        l, b, f_width, f_height = self.figure.get_window_extent().bounds

        # Space needed for labels and ticks
        x_label_space = 0
        if xticks_flag:
            x_label_space += tick_text_size * 1.5
        if xlabel:
            x_label_space += label_text_size * 1.5
        y_label_space = 0
        if yticks_flag:
            y_label_space += ytick_length
        if ylabel:
            y_label_space += label_text_size * 1.5

        ax_plot_rect = (
            float(plot_left_padding + left + y_label_space) / f_width,
            float(plot_bottom_padding + bottom + x_label_space + stats_line_space) / f_height,
            float(width - plot_left_padding - plot_right_padding - y_label_space) / f_width,
            float(
                height
                - plot_bottom_padding
                - plot_top_padding
                - x_label_space
                - plot_title_size
                - 2 * plot_title_padding
                - stats_line_space
            )
            / f_height,
        )
        ax = Axes(self.figure, ax_plot_rect)
        if prefs["square_axis"]:
            l, b, a_width, a_height = ax.get_window_extent().bounds
            delta = abs(a_height - a_width)
            if a_height > a_width:
                a_height = a_width
                ax_plot_rect = (
                    float(plot_left_padding + left) / f_width,
                    float(plot_bottom_padding + bottom + delta / 2.0) / f_height,
                    float(width - plot_left_padding - plot_right_padding) / f_width,
                    float(height - plot_bottom_padding - plot_title_size - 2 * plot_title_padding - delta) / f_height,
                )
            else:
                a_width = a_height
                ax_plot_rect = (
                    float(plot_left_padding + left + delta / 2.0) / f_width,
                    float(plot_bottom_padding + bottom) / f_height,
                    float(width - plot_left_padding - delta) / f_width,
                    float(height - plot_bottom_padding - plot_title_size - 2 * plot_title_padding) / f_height,
                )
            ax.set_position(ax_plot_rect)

        self.figure.add_axes(ax)
        self.ax = ax
        frame = ax.patch
        frame.set_fill(False)

        if frame_flag.lower() == "off":
            self.ax.set_axis_off()
            self.log_xaxis = False
            self.log_yaxis = False
        else:
            # If requested, make x/y axis logarithmic
            if prefs.get("log_xaxis", "False").find("r") >= 0:
                ax.semilogx()
                self.log_xaxis = True
            else:
                self.log_xaxis = False
            if prefs.get("log_yaxis", "False").find("r") >= 0:
                ax.semilogy()
                self.log_yaxis = True
            else:
                self.log_yaxis = False

            if xticks_flag:
                setp(ax.get_xticklabels(), family=prefs["font_family"])
                setp(ax.get_xticklabels(), fontname=prefs["font"])
                setp(ax.get_xticklabels(), size=tick_text_size_point)
            else:
                setp(ax.get_xticklabels(), size=0)

            if yticks_flag:
                setp(ax.get_yticklabels(), family=prefs["font_family"])
                setp(ax.get_yticklabels(), fontname=prefs["font"])
                setp(ax.get_yticklabels(), size=tick_text_size_point)
            else:
                setp(ax.get_yticklabels(), size=0)

            setp(ax.get_xticklines(), markeredgewidth=pixelToPoint(0.5, dpi))
            setp(ax.get_xticklines(), markersize=pixelToPoint(text_size / 2.0, dpi))
            setp(ax.get_yticklines(), markeredgewidth=pixelToPoint(0.5, dpi))
            setp(ax.get_yticklines(), markersize=pixelToPoint(text_size / 2.0, dpi))
            setp(ax.get_xticklines(), zorder=4.0)

            line_width = prefs.get("line_width", 1.0)
            frame_line_width = prefs.get("frame_line_width", line_width)
            grid_line_width = prefs.get("grid_line_width", 0.1)
            plot_line_width = prefs.get("plot_line_width", 0.1)

            setp(ax.patch, linewidth=pixelToPoint(plot_line_width, dpi))
            # setp( ax.spines, linewidth=pixelToPoint(frame_line_width,dpi) )
            # setp( ax.axvline(), linewidth=pixelToPoint(1.0,dpi) )
            axis_grid_flag = prefs.get("plot_axis_grid", True)
            if axis_grid_flag:
                ax.grid(True, color="#555555", linewidth=pixelToPoint(grid_line_width, dpi))

            plot_axis_flag = prefs.get("plot_axis", True)
            if plot_axis_flag:
                # Set labels
                if xlabel:
                    t = ax.set_xlabel(xlabel)
                    t.set_fontname(prefs["font"])
                    t.set_fontsize(label_text_size)

                if ylabel:
                    t = ax.set_ylabel(ylabel)
                    t.set_fontname(prefs["font"])
                    t.set_fontsize(label_text_size)
            else:
                self.ax.set_axis_off()

        # Create a plot title, if necessary
        if plot_title:
            self.ax.title = self.ax.text(
                0.5,
                1.0 + float(plot_title_padding) / height,
                plot_title,
                verticalalignment="bottom",
                horizontalalignment="center",
                size=pixelToPoint(plot_title_size, dpi),
                family=prefs["font_family"],
                fontname=prefs["font"],
            )
            self.ax.title.set_transform(self.ax.transAxes)
            self.ax.title.set_fontname(prefs["font"])
        if stats_line:
            self.ax.stats = self.ax.text(
                0.5,
                (-stats_line_space) / height,
                stats_line,
                verticalalignment="top",
                horizontalalignment="center",
                size=pixelToPoint(stats_line_size, dpi),
            )

            self.ax.stats.set_transform(self.ax.transAxes)



"""Tests for LogFilters."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pytest


from DIRAC.FrameworkSystem.private.standardLogging.LogLevels import LogLevels
from DIRAC.Resources.LogFilters.PatternFilter import PatternFilter
from DIRAC.Resources.LogFilters.ModuleFilter import ModuleFilter, LEVEL


class Record(object):
    def __init__(self, msg, varmessage="", name=None, level=10):
        self.args = (msg,)
        self.varmessage = varmessage
        self.name = name
        self.levelno = level


@pytest.fixture
def pf():
    options = {"Accept": "some,Words", "Reject": "Foo"}
    pf = PatternFilter(options)
    assert pf._accept == ["some", "Words"]
    assert pf._reject == ["Foo"]
    return pf


@pytest.mark.parametrize(
    "record, result",
    [
        (("Some",), False),
        (("some, Words",), True),
        (("some, Words", "Foo"), False),
    ],
)
def test_filter1(pf, record, result):
    assert pf.filter(Record(*record)) == result


@pytest.fixture
def mf():
    options = {"dirac": "ERROR", "l1": "ERROR", "l1.l2": "INFO", "ll1.ll2.ll3": "VERBOSE"}
    mf = ModuleFilter(options)
    assert mf._configDict == {
        "dirac": {LEVEL: LogLevels.ERROR},
        "l1": {LEVEL: LogLevels.ERROR, "l2": {LEVEL: LogLevels.INFO}},
        "ll1": {LEVEL: LogLevels.DEBUG, "ll2": {LEVEL: LogLevels.DEBUG, "ll3": {LEVEL: LogLevels.VERBOSE}}},
    }
    return mf


@pytest.mark.parametrize(
    "name, level, result",
    [
        ("dirac", LogLevels.INFO, False),
        ("dirac", LogLevels.ERROR, True),
        ("l1", LogLevels.INFO, False),
        ("l1", LogLevels.ERROR, True),
        ("l1.ll2", LogLevels.INFO, False),  # inherits from l1
        ("l1.ll2", LogLevels.ERROR, True),  # inherits from l1
        ("l1.l2", LogLevels.INFO, True),  # set to INFO
        ("l1.l2.l3", LogLevels.INFO, True),  # inherits from l1.l2
        ("ll1.ll2", LogLevels.DEBUG, True),  # base level is DEBUG by default
        ("ll1.ll2.ll3", LogLevels.DEBUG, False),  # set to VERBOSE
        ("ll1.ll2.ll3", LogLevels.INFO, True),  # set to VERBOSE
    ],
)
def test_mf(mf, name, level, result):
    assert mf.filter(Record("blabla", name=name, level=level)) == result



""" GraphData encapsulates input data for the DIRAC Graphs plots

    The DIRAC Graphs package is derived from the GraphTool plotting package of the
    CMS/Phedex Project by ... <to be added>
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import six
import time
import datetime
import numpy

from matplotlib.dates import date2num

from DIRAC.Core.Utilities.Graphs.GraphUtilities import to_timestamp, pretty_float

DEBUG = 0


def get_key_type(keys):
    """A utility function to guess the type of the plot keys"""

    min_time_stamp = 1000000000
    max_time_stamp = 1900000000
    time_type = True
    num_type = True
    string_type = True
    key_type = "unknown"
    for key in keys:
        if time_type:
            try:
                time_data = to_timestamp(key)
                if time_data < min_time_stamp or time_data > max_time_stamp:
                    time_type = False
            except ValueError:
                time_type = False
        if num_type:
            try:
                num_data = float(key)
            except Exception:
                num_type = False
        if not isinstance(key, six.string_types):
            string_type = False

    # Take the most restrictive type
    if string_type:
        key_type = "string"
    if num_type:
        key_type = "numeric"
    if time_type:
        key_type = "time"

    return key_type


class GraphData(object):
    def __init__(self, data={}):

        self.truncated = 0
        self.all_keys = []
        self.labels = []
        self.label_values = []
        self.subplots = {}
        self.plotdata = None
        self.data = dict(data)
        self.key_type = "string"
        self.initialize()

    def isEmpty(self):
        """Check if there is no data inserted"""

        return not self.plotdata and not self.subplots

    def setData(self, data):
        """Add data to the GraphData object"""

        self.data = dict(data)
        self.initialize()

    def initialize(self, key_type=None):

        keys = list(self.data)
        if not keys:
            print("GraphData Error: empty data")
        start = time.time()

        if isinstance(self.data[keys[0]], dict):
            for key in self.data:
                self.subplots[key] = PlotData(self.data[key], key_type=key_type)
        else:
            self.plotdata = PlotData(self.data, key_type=key_type)

        if DEBUG:
            print("Time: plot data", time.time() - start, len(self.subplots))

        if self.plotdata:
            self.all_keys = self.plotdata.getKeys()
        else:
            tmpset = set()
            for sub in self.subplots.values():
                for key in sub.getKeys():
                    tmpset.add(key)
            self.all_keys = list(tmpset)

        if key_type:
            self.key_type = key_type
        else:
            self.key_type = get_key_type(self.all_keys)
        self.sortKeys()
        self.makeNumKeys()

        self.sortLabels()

    def expandKeys(self):

        if not self.plotdata:
            for sub in self.subplots:
                self.subplots[sub].expandKeys(self.all_keys)

    def isSimplePlot(self):

        return self.plotdata is not None

    def sortLabels(self, sort_type="max_value", reverse_order=False):
        """Sort labels with a specified method:
        alpha - alphabetic order
        max_value - by max value of the subplot
        sum - by the sum of values of the subplot
        last_value - by the last value in the subplot
        avg_nozeros - by an average that excludes all zero values
        """
        if self.plotdata:
            if self.key_type == "string":
                if sort_type in ["max_value", "sum"]:
                    self.labels = self.plotdata.sortKeys("weight")
                else:
                    self.labels = self.plotdata.sortKeys()
                if reverse_order:
                    self.labels.reverse()
                self.label_values = [self.plotdata.parsed_data[l] for l in self.labels]
        else:
            if sort_type == "max_value":
                pairs = list(zip(list(self.subplots), self.subplots.values()))
                reverse = not reverse_order
                pairs.sort(key=lambda x: x[1].max_value, reverse=reverse)
                self.labels = [x[0] for x in pairs]
                self.label_values = [x[1].max_value for x in pairs]
            elif sort_type == "last_value":
                pairs = list(zip(list(self.subplots), self.subplots.values()))
                reverse = not reverse_order
                pairs.sort(key=lambda x: x[1].last_value, reverse=reverse)
                self.labels = [x[0] for x in pairs]
                self.label_values = [x[1].last_value for x in pairs]
            elif sort_type == "sum":
                pairs = []
                for key in self.subplots:
                    pairs.append((key, self.subplots[key].sum_value))
                reverse = not reverse_order
                pairs.sort(key=lambda x: x[1], reverse=reverse)
                self.labels = [x[0] for x in pairs]
                self.label_values = [x[1] for x in pairs]
            elif sort_type == "alpha":
                self.labels = list(self.subplots)
                self.labels.sort()
                if reverse_order:
                    self.labels.reverse()
                self.label_values = [self.subplots[x].sum_value for x in self.labels]
            elif sort_type == "avg_nozeros":
                pairs = list(zip(list(self.subplots), self.subplots.values()))
                reverse = not reverse_order
                pairs.sort(key=lambda x: x[1].avg_nozeros, reverse=reverse)
                self.labels = [x[0] for x in pairs]
                self.label_values = [x[1].avg_nozeros for x in pairs]
            else:
                self.labels = list(self.subplots)
                if reverse_order:
                    self.labels.reverse()

    def sortKeys(self):
        """Sort the graph keys in a natural order"""

        if self.plotdata:
            self.plotdata.sortKeys()
            self.all_keys = self.plotdata.getKeys()
        else:
            self.all_keys.sort()

        self.min_key = min(self.all_keys)
        self.max_key = max(self.all_keys)

    def makeNumKeys(self):
        """Make numerical representation of the graph keys suitable for plotting"""

        self.all_num_keys = []
        if self.key_type == "string":
            self.all_string_map = {}
            next = 0
            for key in self.all_keys:
                self.all_string_map[key] = next
                self.all_num_keys.append(next)
                next += 1
        elif self.key_type == "time":
            self.all_num_keys = [date2num(datetime.datetime.fromtimestamp(to_timestamp(key))) for key in self.all_keys]
        elif self.key_type == "numeric":
            self.all_num_keys = [float(key) for key in self.all_keys]

        self.min_num_key = min(self.all_num_keys)
        self.max_num_key = max(self.all_num_keys)

    def makeCumulativeGraph(self):
        """Prepare data for the cumulative graph"""

        self.expandKeys()

        if self.plotdata:
            self.plotdata.makeCumulativePlot()
        if self.truncated:
            self.otherPlot.makeCumulativePlot()
        if self.subplots:
            for label in self.subplots:
                self.subplots[label].makeCumulativePlot()

        self.sortLabels(sort_type="last_value")

    def getLabels(self):
        """Get the graph labels together with the numeric values used for the label
        sorting
        """

        labels = []
        if self.plotdata:
            if self.key_type != "string":
                labels = [("NoLabels", 0.0)]
            else:
                labels = list(zip(self.labels, self.label_values))

        elif self.truncated:
            tlabels = self.labels[: self.truncated]
            tvalues = self.label_values[: self.truncated]
            labels = list(zip(tlabels, tvalues))
            labels.append(("Others", sum(self.label_values[self.truncated :])))
        else:
            labels = list(zip(self.labels, self.label_values))

        return labels

    def getStringMap(self):
        """Get string to number mapping for numeric type keys"""
        return self.all_string_map

    def getNumberOfKeys(self):

        return len(self.all_keys)

    def getNumberOfLabels(self):

        if self.truncated:
            return self.truncated + 1
        else:
            return len(self.labels)

    def getPlotNumData(self, label=None, zipFlag=True):
        """Get the plot data in a numeric form"""

        if self.plotdata:
            if zipFlag:
                return zip(self.plotdata.getNumKeys(), self.plotdata.getValues(), self.plotdata.getPlotErrors())
            else:
                return self.plotdata.getValues()
        elif label is not None:
            if label == "Others":
                return self.otherPlot.getPlotDataForNumKeys(self.all_num_keys)
            else:
                return self.subplots[label].getPlotDataForNumKeys(self.all_num_keys)
        else:
            # Get the sum of all the subplots
            self.expandKeys()
            arrays = []
            for label in self.subplots:
                arrays.append(
                    numpy.array([x[1] for x in self.subplots[label].getPlotDataForNumKeys(self.all_num_keys, True)])
                )
            sum_array = sum(arrays)
            if zipFlag:
                return zip(self.all_num_keys, list(sum_array))
            else:
                return sum_array

    def truncateLabels(self, limit=10):
        """Truncate the number of labels to the limit, leave the most important
        ones, accumulate the rest in the 'Others' label
        """

        if self.plotdata:
            return
        nLabels = len(self.labels)
        if nLabels <= limit:
            return

        self.truncated = limit

        new_labels = self.labels[:limit]
        new_labels.append("Others")

        other_data = {}
        for key in self.all_keys:
            other_data[key] = 0.0
        for label in self.labels:
            if label not in new_labels:
                for key in self.all_keys:
                    if key in self.subplots[label].parsed_data:
                        other_data[key] += self.subplots[label].parsed_data[key]
        self.otherPlot = PlotData(other_data)

    def getStats(self):
        """Get statistics of the graph data"""

        numData = self.getPlotNumData(zipFlag=False)
        if not len(numData):  # pylint: disable=len-as-condition
            return 0, 0, 0, 0

        numData = numpy.array(numData)
        min_value = numData.min()
        max_value = numData.max()
        average = float(numData.sum()) / len(numData)
        current = numData[-1]
        return min_value, max_value, average, current

    def getStatString(self, unit=None):
        """Get a string summarizing the graph data statistics"""
        min_value, max_value, average, current = self.getStats()
        tmpList = []
        unitString = ""
        if unit:
            unitString = str(unit)
        if max_value:
            try:
                s = "Max: " + pretty_float(max_value) + " " + unitString
                tmpList.append(s.strip())
            except Exception:
                pass
        if min_value:
            try:
                s = "Min: " + pretty_float(min_value) + " " + unitString
                tmpList.append(s.strip())
            except Exception:
                pass
        if average:
            try:
                s = "Average: " + pretty_float(average) + " " + unitString
                tmpList.append(s.strip())
            except Exception:
                pass
        if current:
            try:
                s = "Current: " + pretty_float(current) + " " + unitString
                tmpList.append(s.strip())
            except Exception:
                pass

        resultString = ", ".join(tmpList)
        return resultString


class PlotData(object):
    """PlotData class is a container for a one dimensional plot data"""

    def __init__(self, data, single=True, key_type=None):

        self.key_type = "unknown"
        if not data:
            print("PlotData Error: empty data")
            return

        # Original data
        self.data = dict(data)

        # Working copy of the parsed data
        self.parsed_data = {}
        self.parsed_errors = {}

        # Keys and values as synchronized lists
        self.keys = []
        self.num_keys = []
        self.values = []
        self.errors = []
        self.sorted_keys = []

        # Do initial data parsing
        self.parseData(key_type)
        if single:
            self.initialize()

    def initialize(self):

        if self.key_type == "string":
            self.keys = self.sortKeys("weight")
        else:
            self.keys = self.sortKeys()

        self.values = [self.parsed_data.get(k, 0.0) for k in self.keys]
        self.errors = [self.parsed_errors.get(k, 0.0) for k in self.keys]
        values_to_sum = [self.parsed_data.get(k, 0.0) for k in self.keys if k != ""]

        self.real_values = []
        for k in self.keys:
            if self.parsed_data[k] is not None:
                self.real_values.append(self.parsed_data[k])

        self.values_sum = float(sum(self.real_values))

        # Prepare numerical representation of keys for plotting
        self.num_keys = []
        if self.key_type == "string":
            self.string_map = {}
            next = 0
            for key in self.keys:
                self.string_map[key] = next
                self.num_keys.append(next)
                next += 1
        elif self.key_type == "time":
            self.num_keys = [date2num(datetime.datetime.fromtimestamp(to_timestamp(key))) for key in self.keys]
        elif self.key_type == "numeric":
            self.num_keys = [float(key) for key in self.keys]

        self.min_value = float(min(self.real_values))
        self.max_value = float(max(self.real_values))
        self.min_key = self.keys[0]
        self.max_key = self.keys[-1]
        self.sum_value = float(sum(self.real_values))
        self.last_value = float(self.real_values[-1])

        count = len(list(filter(lambda a: a != 0, self.real_values)))
        if count != 0:
            self.avg_nozeros = self.sum_value / float(count)
        else:
            self.avg_nozeros = 0

    def expandKeys(self, all_keys):
        """Fill zero values into the missing keys"""

        for k in all_keys:
            if k not in self.parsed_data:
                self.parsed_data[k] = 0.0

        self.sorted_keys = []
        self.keys = list(self.parsed_data)
        self.initialize()

    def sortKeys(self, sort_type="alpha"):
        """Sort keys according to the specified method :
        alpha - sort in alphabetic order
        weight - sort in the order of values
        """

        if self.sorted_keys:
            return self.sorted_keys
        if sort_type == "weight":
            pairs = list(zip(list(self.parsed_data), self.parsed_data.values()))
            pairs.sort(key=lambda x: x[1], reverse=True)
            self.sorted_keys = [x[0] for x in pairs]
        elif sort_type == "alpha":
            self.sorted_keys = list(self.keys)
            self.sorted_keys.sort()
        else:
            print("Unknown sorting type:", sort_type)

        return self.sorted_keys

    def __data_size(self, item):
        """
        Determine a numerical size for the data; this is used to
        sort the keys of the graph.

        If the item is a tuple, take the absolute value of the first entry.
        Otherwise, attempt to take the absolute value of that item.  If that
        fails, just return -1.
        """
        if isinstance(item, tuple):
            return abs(item[0])
        try:
            return abs(item)
        except TypeError:
            return -1

    def parseKey(self, key):
        """
        Parse the name of the pivot; this is the identity function.
        """

        if self.key_type == "time":
            return to_timestamp(key)
        else:
            return key

    def parseDatum(self, data):
        """
        Parse the specific data value; this is the identity.
        """

        if isinstance(data, six.string_types) and "::" in data:
            datum, error = data.split("::")
        elif isinstance(data, tuple):
            datum, error = data
        else:
            error = 0.0
            datum = data

        try:
            resultD = float(datum)
        except Exception:
            resultD = None
        try:
            resultE = float(error)
        except Exception:
            resultE = None

        return (resultD, resultE)

    def parseData(self, key_type=None):
        """
        Parse all the data values passed to the graph.  For this super class,
        basically does nothing except loop through all the data.  A sub-class
        should override the parseDatum and parse_pivot functions rather than
        this one.
        """
        if key_type:
            self.key_type = key_type
        else:
            self.key_type = get_key_type(list(self.data))
        new_parsed_data = {}
        new_passed_errors = {}
        for key, data in self.data.items():
            new_key = self.parseKey(key)
            data, error = self.parseDatum(data)
            # if data != None:
            new_parsed_data[new_key] = data
            new_passed_errors[new_key] = error
        self.parsed_data = new_parsed_data
        self.parsed_errors = new_passed_errors

        self.keys = list(self.parsed_data)

    def makeCumulativePlot(self):

        if not self.sorted_keys:
            self.sortKeys()

        cum_values = []
        if self.values[0] is None:
            cum_values.append(0.0)
        else:
            cum_values.append(self.values[0])
        for i in range(1, len(self.values)):
            if self.values[i] is None:
                cum_values.append(cum_values[i - 1])
            else:
                cum_values.append(cum_values[i - 1] + self.values[i])

        self.values = cum_values
        self.last_value = float(self.values[-1])

    def getPlotData(self):

        return self.parsed_data

    def getPlotErrors(self):

        return self.parsed_errors

    def getPlotNumData(self):

        return zip(self.num_keys, self.values, self.errors)

    def getPlotDataForKeys(self, keys):

        result_pairs = []
        for key in keys:
            if key in self.parsed_data:
                result_pairs.append(key, self.parsed_data[key], self.parsed_errors[key])
            else:
                result_pairs.append(key, None, 0.0)

        return result_pairs

    def getPlotDataForNumKeys(self, num_keys, zeroes=False):

        result_pairs = []
        for num_key in num_keys:
            try:
                ind = self.num_keys.index(num_key)
                if self.values[ind] is None and zeroes:
                    result_pairs.append((self.num_keys[ind], 0.0, 0.0))
                else:
                    result_pairs.append((self.num_keys[ind], self.values[ind], self.errors[ind]))
            except ValueError:
                if zeroes:
                    result_pairs.append((num_key, 0.0, 0.0))
                else:
                    result_pairs.append((num_key, None, 0.0))

        return result_pairs

    def getKeys(self):

        return self.keys

    def getNumKeys(self):

        return self.num_keys

    def getValues(self):

        return self.values

    def getMaxValue(self):

        return max(self.values)

    def getMinValue(self):

        return min(self.values)



#!/usr/bin/env python
"""
Add an existing transformation to an existing production.
Transformations already belonging to another production cannot be added.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("prodID:         Production ID")
    Script.registerArgument("transID:        Transformation ID")
    Script.registerArgument("parentTransID:  Parent Transformation ID", default="", mandatory=False)
    _, args = Script.parseCommandLine()

    from DIRAC.ProductionSystem.Client.ProductionClient import ProductionClient
    from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

    prodClient = ProductionClient()
    transClient = TransformationClient()

    # get arguments
    prodID, transID, parentTransID = Script.getPositionalArgs(group=True)
    if len(args) > 3:
        Script.showHelp(exitCode=1)

    res = transClient.getTransformation(transID)
    if not res["OK"]:
        DIRAC.gLogger.error("Failed to get transformation %s: %s" % (transID, res["Message"]))
        DIRAC.exit(-1)

    transID = res["Value"]["TransformationID"]

    if parentTransID:
        res = transClient.getTransformation(parentTransID)
        if not res["OK"]:
            DIRAC.gLogger.error("Failed to get transformation %s: %s" % (parentTransID, res["Message"]))
            DIRAC.exit(-1)
        parentTransID = res["Value"]["TransformationID"]

    res = prodClient.getProduction(prodID)
    if not res["OK"]:
        DIRAC.gLogger.error("Failed to get production %s: %s" % (prodID, res["Message"]))
        DIRAC.exit(-1)

    prodID = res["Value"]["ProductionID"]
    res = prodClient.addTransformationsToProduction(prodID, transID, parentTransID)
    if not res["OK"]:
        DIRAC.gLogger.error(res["Message"])
        DIRAC.exit(-1)

    if parentTransID:
        msg = "Transformation %s successfully added to production %s with parent transformation %s" % (
            transID,
            prodID,
            parentTransID,
        )
    else:
        msg = "Transformation %s successfully added to production %s with no parent transformation" % (transID, prodID)

    DIRAC.gLogger.notice(msg)

    DIRAC.exit(0)


if __name__ == "__main__":
    main()



""" VirtualMachineHandler provides remote access to VirtualMachineDB

    The following methods are available in the Service interface:

    - insertInstance
    - declareInstanceSubmitted
    - declareInstanceRunning
    - instanceIDHeartBeat
    - declareInstanceHalting
    - getInstancesByStatus
    - declareInstancesStopping
    - getUniqueID( instanceID ) return cloud manager uniqueID form VMDIRAC instanceID

"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
from subprocess import Popen, PIPE
import six

from DIRAC import gLogger, S_ERROR, S_OK
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.WorkloadManagementSystem.DB.VirtualMachineDB import VirtualMachineDB
from DIRAC.Core.Security.Properties import OPERATOR, VM_RPC_OPERATION
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getVMTypeConfig, getVMTypes
from DIRAC.Resources.Cloud.Utilities import STATE_MAP
from DIRAC.Resources.Cloud.EndpointFactory import EndpointFactory
from DIRAC.WorkloadManagementSystem.Utilities.Utils import getProxyFileForCloud


class VirtualMachineManagerHandler(RequestHandler):
    @classmethod
    def initializeHandler(cls, serviceInfo):
        cls.virtualMachineDB = VirtualMachineDB()
        cls.haltStalledInstances()
        cls.checkStalledInstances()

        if cls.virtualMachineDB._connected:
            gThreadScheduler.addPeriodicTask(60 * 15, cls.checkStalledInstances)
            return S_OK()
        return S_ERROR()

    @classmethod
    def haltStalledInstances(cls):
        result = cls.virtualMachineDB.getInstancesByStatus("Stalled")
        if not result["OK"]:
            return result

        uList = []
        for image in result["Value"]:
            uList += result["Value"][image]

        stallingList = []
        for uID in uList:
            result = cls.virtualMachineDB.getInstanceID(uID)
            if not result["OK"]:
                continue
            stallingList.append(result["Value"])
        return cls.haltInstances(stallingList)

    @classmethod
    def getCEInstances(cls, siteList=None, ceList=None, vo=None):

        result = getVMTypes(siteList=siteList, ceList=ceList, vo=vo)
        if not result["OK"]:
            return result
        imageDict = result["Value"]
        ceList = []
        for site in imageDict:
            for ce in imageDict[site]:
                result = EndpointFactory().getCE(site, ce)
                if not result["OK"]:
                    continue
                ceList.append((site, ce, result["Value"]))

        nodeDict = {}
        for site, ceName, ce in ceList:
            result = ce.getVMNodes()
            if not result["OK"]:
                continue
            for node in result["Value"]:
                if not node.name.startswith("DIRAC"):
                    continue
                ip = node.public_ips[0] if node.public_ips else "None"
                nodeState = (
                    node.state.upper() if not isinstance(node.state, six.integer_types) else STATE_MAP[node.state]
                )
                nodeDict[node.id] = {
                    "Site": site,
                    "CEName": ceName,
                    "NodeName": node.name,
                    "PublicIP": ip,
                    "State": nodeState,
                }
        return S_OK(nodeDict)

    @classmethod
    def checkStalledInstances(cls):
        """
        To avoid stalling instances consuming resources at cloud endpoint,
        attempts to halt the stalled list in the cloud endpoint
        """
        result = cls.virtualMachineDB.declareStalledInstances()
        if not result["OK"]:
            return result

        stallingList = result["Value"]
        return cls.haltInstances(stallingList)

    @classmethod
    def stopInstance(cls, site, endpoint, nodeID):

        result = getVMTypeConfig(site, endpoint)
        if not result["OK"]:
            return result
        ceParams = result["Value"]
        ceFactory = EndpointFactory()
        result = ceFactory.getCEObject(parameters=ceParams)
        if not result["OK"]:
            return result

        ce = result["Value"]
        return ce.stopVM(nodeID)

    @classmethod
    def createEndpoint(cls, uniqueID):

        result = cls.virtualMachineDB.getEndpointFromInstance(uniqueID)
        if not result["OK"]:
            return result
        site, endpoint = result["Value"].split("::")

        result = getVMTypeConfig(site, endpoint)
        if not result["OK"]:
            return result
        ceParams = result["Value"]
        ceFactory = EndpointFactory()
        return ceFactory.getCEObject(parameters=ceParams)

    @classmethod
    def haltInstances(cls, vmList):
        """
        Common haltInstances for Running(from class VirtualMachineManagerHandler) and
        Stalled(from checkStalledInstances periodic task) to Halt
        """
        failed = {}
        successful = {}

        for instanceID in vmList:
            instanceID = int(instanceID)
            result = cls.virtualMachineDB.getUniqueID(instanceID)
            if not result["OK"]:
                gLogger.error("haltInstances: on getUniqueID call: %s" % result["Message"])
                continue
            uniqueID = result["Value"]

            result = cls.createEndpoint(uniqueID)
            if not result["OK"]:
                gLogger.error("haltInstances: on createEndpoint call: %s" % result["Message"])
                continue

            endpoint = result["Value"]

            # Get proxy to be used to connect to the cloud endpoint
            authType = endpoint.parameters.get("Auth")
            if authType and authType.lower() in ["x509", "voms"]:
                siteName = endpoint.parameters["Site"]
                ceName = endpoint.parameters["CEName"]
                gLogger.verbose("Getting cloud proxy for %s/%s" % (siteName, ceName))
                result = getProxyFileForCloud(endpoint)
                if not result["OK"]:
                    continue
                endpoint.setProxy(result["Value"])

            result = endpoint.stopVM(uniqueID)
            if result["OK"]:
                cls.virtualMachineDB.recordDBHalt(instanceID, 0)
                successful[instanceID] = True
            else:
                failed[instanceID] = result["Message"]

        return S_OK({"Successful": successful, "Failed": failed})

    @classmethod
    def getPilotOutput(cls, pilotRef):
        if not pilotRef.startswith("vm://"):
            return S_ERROR("Invalid pilot reference %s" % pilotRef)

        # Get the VM public IP
        diracID, nPilot = os.path.basename(pilotRef).split(":")
        result = cls.virtualMachineDB.getUniqueIDByName(diracID)
        if not result["OK"]:
            return result
        uniqueID = result["Value"]
        result = cls.virtualMachineDB.getInstanceID(uniqueID)
        if not result["OK"]:
            return result
        instanceID = result["Value"]
        result = cls.virtualMachineDB.getInstanceParameter("PublicIP", instanceID)
        if not result["OK"]:
            return result
        publicIP = result["Value"]

        op = Operations()
        privateKeyFile = op.getValue("/Cloud/PrivateKey", "")
        diracUser = op.getValue("/Cloud/VMUser", "")

        ssh_str = "%s@%s" % (diracUser, publicIP)
        cmd = ["ssh", "-i", privateKeyFile, ssh_str, "cat /etc/joboutputs/vm-pilot.%s.log" % nPilot]
        inst = Popen(cmd, stdout=PIPE, stderr=PIPE, stdin=PIPE)
        output, stderr = inst.communicate()
        if inst.returncode:
            return S_ERROR("Failed to get pilot output: %s" % stderr)
        else:
            return S_OK(output)

    def initialize(self):
        credDict = self.getRemoteCredentials()
        self.rpcProperties = credDict["properties"]

    types_getCEInstances = [(list, type(None)), (list, type(None)), six.string_types]

    def export_getCEInstances(self, siteList, ceList, vo):

        if not siteList:
            siteList = None
        return self.getCEInstances(siteList=siteList, ceList=ceList, vo=vo)

    types_stopInstance = [six.string_types, six.string_types, six.string_types]

    def export_stopInstance(self, site, endpoint, nodeID):

        return self.stopInstance(site, endpoint, nodeID)

    types_getPilotOutput = [six.string_types]

    def export_getPilotOutput(self, pilotReference):

        return self.getPilotOutput(pilotReference)

    types_checkVmWebOperation = [six.string_types]

    def export_checkVmWebOperation(self, operation):
        """
        return true if rpc has OPERATOR
        """
        if OPERATOR in self.rpcProperties:
            return S_OK("Auth")
        return S_OK("Unauth")

    types_insertInstance = [six.string_types, six.string_types, six.string_types, six.string_types, six.string_types]

    def export_insertInstance(self, uniqueID, imageName, instanceName, endpoint, runningPodName):
        """
        Check Status of a given image
        Will insert a new Instance in the DB
        """
        return self.virtualMachineDB.insertInstance(uniqueID, imageName, instanceName, endpoint, runningPodName)

    types_getUniqueID = [six.string_types]

    def export_getUniqueID(self, instanceID):
        """
        return cloud manager uniqueID from VMDIRAC instanceID
        """
        return self.virtualMachineDB.getUniqueID(instanceID)

    types_getUniqueIDByName = [six.string_types]

    def export_getUniqueIDByName(self, instanceName):
        """
        return cloud manager uniqueID from VMDIRAC name
        """
        return self.virtualMachineDB.getUniqueIDByName(instanceName)

    types_setInstanceUniqueID = [six.integer_types, six.string_types]

    def export_setInstanceUniqueID(self, instanceID, uniqueID):
        """
        Check Status of a given image
        Will insert a new Instance in the DB
        """
        return self.virtualMachineDB.setInstanceUniqueID(instanceID, uniqueID)

    types_declareInstanceSubmitted = [six.string_types]

    def export_declareInstanceSubmitted(self, uniqueID):
        """
        After submission of the instance the Director should declare the new Status
        """
        return self.virtualMachineDB.declareInstanceSubmitted(uniqueID)

    types_declareInstanceRunning = [six.string_types, six.string_types]

    def export_declareInstanceRunning(self, uniqueID, privateIP):
        """
        Declares an instance Running and sets its associated info (uniqueID, publicIP, privateIP)
        Returns S_ERROR if:
        - instanceName does not have a "Submitted" entry
        - uniqueID is not unique
        """
        gLogger.info("Declare instance Running uniqueID: %s" % (uniqueID))
        publicIP = self.getRemoteAddress()[0]
        gLogger.info("Declare instance Running publicIP: %s" % (publicIP))

        return self.virtualMachineDB.declareInstanceRunning(uniqueID, publicIP, privateIP)

    types_instanceIDHeartBeat = [six.string_types, float, six.integer_types, six.integer_types, six.integer_types]

    def export_instanceIDHeartBeat(self, uniqueID, load, jobs, transferredFiles, transferredBytes, uptime=0):
        """
        Insert the heart beat info from a running instance
        It checks the status of the instance and the corresponding image
        Declares "Running" the instance and the image
        It returns S_ERROR if the status is not OK
        """
        try:
            uptime = int(uptime)
        except ValueError:
            uptime = 0

        return self.virtualMachineDB.instanceIDHeartBeat(
            uniqueID, load, jobs, transferredFiles, transferredBytes, uptime
        )

    types_declareInstancesStopping = [list]

    def export_declareInstancesStopping(self, instanceIdList):
        """
        Declares "Stopping" the instance because the Delete button of Browse Instances
        The instanceID is the VMDIRAC VM id
        When next instanceID heat beat with stopping status on the DB the VM will stop the job agent and terminates properly
        It returns S_ERROR if the status is not OK
        """
        for instanceID in instanceIdList:
            gLogger.info("Stopping DIRAC instanceID: %s" % (instanceID))
            result = self.virtualMachineDB.getInstanceStatus(instanceID)
            if not result["OK"]:
                return result
            state = result["Value"]
            gLogger.info("Stopping DIRAC instanceID: %s, current state %s" % (instanceID, state))

            if state == "Stalled":
                result = self.virtualMachineDB.getUniqueID(instanceID)
                if not result["OK"]:
                    return result
                uniqueID = result["Value"]
                result = self.export_declareInstanceHalting(uniqueID, 0)
                if not result["OK"]:
                    return result
            elif state == "New":
                result = self.virtualMachineDB.recordDBHalt(instanceID, 0)
                if not result["OK"]:
                    return result
            else:
                # this is only applied to allowed transitions
                result = self.virtualMachineDB.declareInstanceStopping(instanceID)
                if not result["OK"]:
                    return result
        return S_OK()

    types_declareInstanceHalting = [six.string_types, float]

    def export_declareInstanceHalting(self, uniqueID, load):
        """
        Insert the heart beat info from a halting instance
        The VM has the uniqueID, which is the Cloud manager VM id
        Declares "Halted" the instance and the image
        It returns S_ERROR if the status is not OK
        """
        result = self.virtualMachineDB.declareInstanceHalting(uniqueID, load)
        if not result["OK"]:
            if "Halted ->" not in result["Message"]:
                return result
            else:
                gLogger.info("Bad transition from Halted to something, will assume Halted")

        haltingList = []
        instanceID = self.virtualMachineDB.getInstanceID(uniqueID)
        if not instanceID["OK"]:
            return instanceID
        instanceID = instanceID["Value"]
        haltingList.append(instanceID)

        return self.haltInstances(haltingList)

    types_getInstancesByStatus = [six.string_types]

    def export_getInstancesByStatus(self, status):
        """
        Get dictionary of Image Names with InstanceIDs in given status
        """
        return self.virtualMachineDB.getInstancesByStatus(status)

    types_getAllInfoForUniqueID = [six.string_types]

    def export_getAllInfoForUniqueID(self, uniqueID):
        """
        Get all the info for a UniqueID
        """
        return self.virtualMachineDB.getAllInfoForUniqueID(uniqueID)

    types_getInstancesContent = [dict, (list, tuple), six.integer_types, six.integer_types]

    def export_getInstancesContent(self, selDict, sortDict, start, limit):
        """
        Retrieve the contents of the DB
        """
        return self.virtualMachineDB.getInstancesContent(selDict, sortDict, start, limit)

    types_getHistoryForInstanceID = [six.integer_types]

    def export_getHistoryForInstanceID(self, instanceId):
        """
        Retrieve the contents of the DB
        """
        return self.virtualMachineDB.getHistoryForInstanceID(instanceId)

    types_getInstanceCounters = [six.string_types, dict]

    def export_getInstanceCounters(self, groupField, selDict):
        """
        Retrieve the contents of the DB
        """
        return self.virtualMachineDB.getInstanceCounters(groupField, selDict)

    types_getHistoryValues = [int, dict]

    def export_getHistoryValues(self, averageBucket, selDict, fields2Get=None, timespan=0):
        """
        Retrieve the contents of the DB
        """
        if not fields2Get:
            fields2Get = []
        return self.virtualMachineDB.getHistoryValues(averageBucket, selDict, fields2Get, timespan)

    types_getRunningInstancesHistory = [int, int]

    def export_getRunningInstancesHistory(self, timespan, bucketSize):
        """
        Retrieve number of running instances in each bucket
        """
        return self.virtualMachineDB.getRunningInstancesHistory(timespan, bucketSize)

    types_getRunningInstancesBEPHistory = [int, int]

    def export_getRunningInstancesBEPHistory(self, timespan, bucketSize):
        """
        Retrieve number of running instances in each bucket by End-Point History
        """
        return self.virtualMachineDB.getRunningInstancesBEPHistory(timespan, bucketSize)

    types_getRunningInstancesByRunningPodHistory = [int, int]

    def export_getRunningInstancesByRunningPodHistory(self, timespan, bucketSize):
        """
        Retrieve number of running instances in each bucket by Running Pod History
        """
        return self.virtualMachineDB.getRunningInstancesByRunningPodHistory(timespan, bucketSize)

    types_getRunningInstancesByImageHistory = [int, int]

    def export_getRunningInstancesByImageHistory(self, timespan, bucketSize):
        """
        Retrieve number of running instances in each bucket by Running Pod History
        """
        return self.virtualMachineDB.getRunningInstancesByImageHistory(timespan, bucketSize)



#!/usr/bin/env python
########################################################################
# File :   dirac_admin_accounting_cli
# Author : Adria Casajus
########################################################################
"""
Command line administrative interface to DIRAC Accounting DataStore Service
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.localCfg.addDefaultEntry("LogLevel", "info")
    Script.parseCommandLine()

    from DIRAC.AccountingSystem.Client.AccountingCLI import AccountingCLI

    acli = AccountingCLI()
    acli.start()


if __name__ == "__main__":
    main()



import six
import time
import json

from DIRAC import S_OK, S_ERROR, gLogger

from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getDNForUsername
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.Core.Utilities.JEncode import decode

from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.RequestManagementSystem.private.RequestValidator import RequestValidator
from DIRAC.TransformationSystem.Client import TransformationFilesStatus
from DIRAC.TransformationSystem.Client.BodyPlugin.BaseBody import BaseBody
from DIRAC.TransformationSystem.Client.TaskManager import TaskBase


class StopTaskIteration(Exception):
    """Utility Exception to stop creating
    a Request for the current worked on task
    """

    pass


class RequestTasks(TaskBase):
    """
    Class for handling tasks for the RMS
    """

    def __init__(
        self,
        transClient=None,
        logger=None,
        requestClient=None,
        requestClass=None,
        requestValidator=None,
        ownerDN=None,
        ownerGroup=None,
    ):
        """c'tor

        the requestClass is by default Request.
        If extensions want to use an extended type, they can pass it as a parameter.
        This is the same behavior as WorfkloTasks and jobClass
        """

        if not logger:
            logger = gLogger.getSubLogger("RequestTasks")

        super(RequestTasks, self).__init__(transClient, logger)
        useCertificates = True if (bool(ownerDN) and bool(ownerGroup)) else False

        if not requestClient:
            self.requestClient = ReqClient(
                useCertificates=useCertificates, delegatedDN=ownerDN, delegatedGroup=ownerGroup
            )
        else:
            self.requestClient = requestClient

        if not requestClass:
            self.requestClass = Request
        else:
            self.requestClass = requestClass

        if not requestValidator:
            self.requestValidator = RequestValidator()
        else:
            self.requestValidator = requestValidator

    def prepareTransformationTasks(
        self, transBody, taskDict, owner="", ownerGroup="", ownerDN="", bulkSubmissionFlag=False
    ):
        """Prepare tasks, given a taskDict, that is created (with some manipulation) by the DB"""
        if not taskDict:
            return S_OK({})

        if (not owner) or (not ownerGroup):
            res = getProxyInfo(False, False)
            if not res["OK"]:
                return res
            proxyInfo = res["Value"]
            owner = proxyInfo["username"]
            ownerGroup = proxyInfo["group"]

        if not ownerDN:
            res = getDNForUsername(owner)
            if not res["OK"]:
                return res
            ownerDN = res["Value"][0]

        try:
            transJson, _decLen = decode(transBody)

            if isinstance(transJson, BaseBody):
                self._bodyPlugins(transJson, taskDict, ownerDN, ownerGroup)
            else:
                self._multiOperationsBody(transJson, taskDict, ownerDN, ownerGroup)
        except ValueError:  # #json couldn't load
            self._singleOperationsBody(transBody, taskDict, ownerDN, ownerGroup)

        return S_OK(taskDict)

    def _multiOperationsBody(self, transJson, taskDict, ownerDN, ownerGroup):
        """Deal with a Request that has multiple operations

        :param transJson: list of lists of string and dictionaries, e.g.:

          .. code :: python

            body = [ ( "ReplicateAndRegister", { "SourceSE":"FOO-SRM", "TargetSE":"TASK:TargetSE" }),
                     ( "RemoveReplica", { "TargetSE":"FOO-SRM" } ),
                   ]

            If a value of an operation parameter in the body starts with ``TASK:``,
            we take it from the taskDict.
            For example ``TASK:TargetSE`` is replaced with ``task['TargetSE']``

        :param dict taskDict: dictionary of tasks, modified in this function
        :param str ownerDN: certificate DN used for the requests
        :param str onwerGroup: dirac group used for the requests

        :returns: None
        """
        for taskID, task in list(taskDict.items()):
            try:
                transID = task["TransformationID"]
                if not task.get("InputData"):
                    raise StopTaskIteration("No input data")
                files = []

                oRequest = Request()
                if isinstance(task["InputData"], list):
                    files = task["InputData"]
                elif isinstance(task["InputData"], six.string_types):
                    files = task["InputData"].split(";")

                # create the operations from the json structure
                for operationTuple in transJson:
                    op = Operation()
                    op.Type = operationTuple[0]
                    for parameter, value in operationTuple[1].items():
                        # Here we massage a bit the body to replace some parameters
                        # with what we have in the task.
                        try:
                            taskKey = value.split("TASK:")[1]
                            value = task[taskKey]
                        # Either the attribute is not a string (AttributeError)
                        # or it does not start with 'TASK:' (IndexError)
                        except (AttributeError, IndexError):
                            pass
                        # That happens when the requested substitution is not
                        # a key in the task, and that's a problem
                        except KeyError:
                            raise StopTaskIteration("Parameter %s does not exist in taskDict" % taskKey)

                        setattr(op, parameter, value)

                    for lfn in files:
                        opFile = File()
                        opFile.LFN = lfn
                        op.addFile(opFile)

                    oRequest.addOperation(op)

                result = self._assignRequestToTask(oRequest, taskDict, transID, taskID, ownerDN, ownerGroup)
                if not result["OK"]:
                    raise StopTaskIteration("Could not assign request to task: %s" % result["Message"])
            except StopTaskIteration as e:
                self._logError("Error creating request for task", "%s, %s" % (taskID, e), transID=transID)
                taskDict.pop(taskID)

    def _singleOperationsBody(self, transBody, taskDict, ownerDN, ownerGroup):
        """deal with a Request that has just one operation, as it was sofar

        :param transBody: string, can be an empty string
        :param dict taskDict: dictionary of tasks, modified in this function
        :param str ownerDN: certificate DN used for the requests
        :param str onwerGroup: dirac group used for the requests

        :returns: None
        """

        requestOperation = "ReplicateAndRegister"
        if transBody:
            try:
                _requestType, requestOperation = transBody.split(";")
            except AttributeError:
                pass
        failedTasks = []
        # Do not remove sorted, we might pop elements in the loop
        for taskID, task in taskDict.items():

            transID = task["TransformationID"]

            oRequest = Request()
            transfer = Operation()
            transfer.Type = requestOperation
            transfer.TargetSE = task["TargetSE"]

            # If there are input files
            if task.get("InputData"):
                if isinstance(task["InputData"], list):
                    files = task["InputData"]
                elif isinstance(task["InputData"], six.string_types):
                    files = task["InputData"].split(";")
                for lfn in files:
                    trFile = File()
                    trFile.LFN = lfn

                    transfer.addFile(trFile)

            oRequest.addOperation(transfer)
            result = self._assignRequestToTask(oRequest, taskDict, transID, taskID, ownerDN, ownerGroup)
            if not result["OK"]:
                failedTasks.append(taskID)
        # Remove failed tasks
        for taskID in failedTasks:
            taskDict.pop(taskID)

    def _bodyPlugins(self, bodyObj, taskDict, ownerDN, ownerGroup):
        """Deal with complex body object"""
        for taskID, task in list(taskDict.items()):
            try:
                transID = task["TransformationID"]
                if not task.get("InputData"):
                    raise StopTaskIteration("No input data")

                oRequest = bodyObj.taskToRequest(taskID, task, transID)
                result = self._assignRequestToTask(oRequest, taskDict, transID, taskID, ownerDN, ownerGroup)
                if not result["OK"]:
                    raise StopTaskIteration("Could not assign request to task: %s" % result["Message"])
            except StopTaskIteration as e:
                self._logError("Error creating request for task", "%s, %s" % (taskID, e), transID=transID)
                taskDict.pop(taskID)

    def _assignRequestToTask(self, oRequest, taskDict, transID, taskID, ownerDN, ownerGroup):
        """set ownerDN and group to request, and add the request to taskDict if it is
        valid, otherwise remove the task from the taskDict

        :param oRequest: Request
        :param dict taskDict: dictionary of tasks, modified in this function
        :param int transID: Transformation ID
        :param int taskID: Task ID
        :param str ownerDN: certificate DN used for the requests
        :param str onwerGroup: dirac group used for the requests

        :returns: None
        """

        oRequest.RequestName = self._transTaskName(transID, taskID)
        oRequest.OwnerDN = ownerDN
        oRequest.OwnerGroup = ownerGroup

        isValid = self.requestValidator.validate(oRequest)
        if not isValid["OK"]:
            self._logError("Error creating request for task", "%s %s" % (taskID, isValid), transID=transID)
            return S_ERROR("Error creating request")
        taskDict[taskID]["TaskObject"] = oRequest
        return S_OK()

    def submitTransformationTasks(self, taskDict):
        """Submit requests one by one"""
        submitted = 0
        failed = 0
        startTime = time.time()
        method = "submitTransformationTasks"
        for task in taskDict.values():
            # transID is the same for all tasks, so pick it up every time here
            transID = task["TransformationID"]
            if not task["TaskObject"]:
                task["Success"] = False
                failed += 1
                continue
            res = self.submitTaskToExternal(task["TaskObject"])
            if res["OK"]:
                task["ExternalID"] = res["Value"]
                task["Success"] = True
                submitted += 1
            else:
                self._logError("Failed to submit task to RMS", res["Message"], transID=transID)
                task["Success"] = False
                failed += 1
        if submitted:
            self._logInfo(
                "Submitted %d tasks to RMS in %.1f seconds" % (submitted, time.time() - startTime),
                transID=transID,
                method=method,
            )
        if failed:
            self._logWarn("Failed to submit %d tasks to RMS." % (failed), transID=transID, method=method)
        return S_OK(taskDict)

    def submitTaskToExternal(self, oRequest):
        """
        Submits a request to RMS
        """
        if isinstance(oRequest, self.requestClass):
            return self.requestClient.putRequest(oRequest, useFailoverProxy=False, retryMainService=2)
        return S_ERROR("Request should be a Request object")

    def updateTransformationReservedTasks(self, taskDicts):
        requestNameIDs = {}
        noTasks = []
        for taskDict in taskDicts:
            requestName = self._transTaskName(taskDict["TransformationID"], taskDict["TaskID"])
            reqID = taskDict["ExternalID"]
            if reqID and int(reqID):
                requestNameIDs[requestName] = reqID
            else:
                noTasks.append(requestName)
        return S_OK({"NoTasks": noTasks, "TaskNameIDs": requestNameIDs})

    def getSubmittedTaskStatus(self, taskDicts):
        """
        Check if tasks changed status, and return a list of tasks per new status
        """
        updateDict = {}
        badRequestID = 0
        for taskDict in taskDicts:
            oldStatus = taskDict["ExternalStatus"]
            # ExternalID is normally a string
            if taskDict["ExternalID"] and int(taskDict["ExternalID"]):
                newStatus = self.requestClient.getRequestStatus(taskDict["ExternalID"])
                if not newStatus["OK"]:
                    log = self._logVerbose if "not exist" in newStatus["Message"] else self._logWarn
                    log(
                        "getSubmittedTaskStatus: Failed to get requestID for request",
                        newStatus["Message"],
                        transID=taskDict["TransformationID"],
                    )
                else:
                    newStatus = newStatus["Value"]
                    # We don't care updating the tasks to Assigned while the request is being processed
                    if newStatus != oldStatus and newStatus != "Assigned":
                        updateDict.setdefault(newStatus, []).append(taskDict["TaskID"])
            else:
                badRequestID += 1
        if badRequestID:
            self._logWarn("%d requests have identifier 0" % badRequestID)
        return S_OK(updateDict)

    def getSubmittedFileStatus(self, fileDicts):
        """
        Check if transformation files changed status, and return a list of taskIDs per new status
        """
        # Don't try and get status of not submitted tasks!
        transID = None
        taskFiles = {}
        for fileDict in fileDicts:
            # There is only one transformation involved, get however the transID in the loop
            transID = fileDict["TransformationID"]
            taskID = int(fileDict["TaskID"])
            taskFiles.setdefault(taskID, []).append(fileDict["LFN"])
        # Should not happen, but just in case there are no files, return
        if transID is None:
            return S_OK({})

        res = self.transClient.getTransformationTasks({"TransformationID": transID, "TaskID": list(taskFiles)})
        if not res["OK"]:
            return res
        requestFiles = {}
        for taskDict in res["Value"]:
            taskID = taskDict["TaskID"]
            externalID = taskDict["ExternalID"]
            # Only consider tasks that are submitted, ExternalID is a string
            if taskDict["ExternalStatus"] != "Created" and externalID and int(externalID):
                requestFiles[externalID] = taskFiles[taskID]

        updateDict = {}
        for requestID, lfnList in requestFiles.items():
            statusDict = self.requestClient.getRequestFileStatus(requestID, lfnList)
            if not statusDict["OK"]:
                log = self._logVerbose if "not exist" in statusDict["Message"] else self._logWarn
                log(
                    "Failed to get files status for request",
                    statusDict["Message"],
                    transID=transID,
                    method="getSubmittedFileStatus",
                )
            else:
                for lfn, newStatus in statusDict["Value"].items():
                    if newStatus == "Done":
                        updateDict[lfn] = TransformationFilesStatus.PROCESSED
                    elif newStatus == "Failed":
                        updateDict[lfn] = TransformationFilesStatus.PROBLEMATIC
        return S_OK(updateDict)



"""
Just a selector for X509Certificate
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


from pkgutil import extend_path


#####
# SUPER DISGUSTING HACK
# We define these variables, and then remove them immediately.
# it is to allow something like 'from DIRAC.Core.Security import X509Chain'
# But pylint would complain just like that
# I've spent a lot of time trying to get pylint to work, but...
# https://github.com/PyCQA/pylint/issues/2474

X509Chain = None
X509CRL = None
X509Certificate = None
X509Request = None

locals().pop("X509Chain")
locals().pop("X509CRL")
locals().pop("X509Certificate")
locals().pop("X509Request")
####


# Add the m2crypto subpackage to the search path
# This allows imports like 'from DIRAC.Core.Security.X509Chian...' to work transparently
# Nice kind of tricks you find in libraries like xml...
__path__ = extend_path(__path__, __name__ + ".m2crypto")



#!/usr/bin/env python
"""
Get the transformations belonging to a given production

Example:
  $ dirac-prod-get-trans 381
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.PrettyPrint import printTable
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("prodID: Production ID")
    _, args = Script.parseCommandLine()

    from DIRAC.ProductionSystem.Client.ProductionClient import ProductionClient
    from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

    # get arguments
    prodID = args[0]

    prodClient = ProductionClient()
    transClient = TransformationClient()

    res = prodClient.getProductionTransformations(prodID)
    transIDs = []

    if res["OK"]:
        transList = res["Value"]
        if not transList:
            DIRAC.gLogger.notice("No transformation associated with production %s" % prodID)
            DIRAC.exit(-1)
        for trans in transList:
            transIDs.append(trans["TransformationID"])
    else:
        DIRAC.gLogger.error(res["Message"])
        DIRAC.exit(-1)

    fields = [
        "TransformationName",
        "Status",
        "F_Proc.",
        "F_Proc.(%)",
        "TransformationID",
        "ProductionID",
        "Prod_LastUpdate",
        "Prod_InsertedTime",
    ]

    records = []

    paramShowNames = [
        "TransformationID",
        "TransformationName",
        "Type",
        "Status",
        "Files_Total",
        "Files_PercentProcessed",
        "Files_Processed",
        "Files_Unused",
        "Jobs_TotalCreated",
        "Jobs_Waiting",
        "Jobs_Running",
        "Jobs_Done",
        "Jobs_Failed",
        "Jobs_Stalled",
    ]
    resList = []

    res = transClient.getTransformationSummaryWeb({"TransformationID": transIDs}, [], 0, len(transIDs))

    if not res["OK"]:
        DIRAC.gLogger.error(res["Message"])
        DIRAC.exit(-1)

    if res["Value"]["TotalRecords"] > 0:
        paramNames = res["Value"]["ParameterNames"]
        for paramValues in res["Value"]["Records"]:
            paramShowValues = map(lambda pname: paramValues[paramNames.index(pname)], paramShowNames)
            showDict = dict(zip(paramShowNames, paramShowValues))
            resList.append(showDict)

    for res in resList:
        files_Processed = res["Files_Processed"]
        files_PercentProcessed = res["Files_PercentProcessed"]
        status = res["Status"]
        type = res["Type"]
        transName = res["TransformationName"]
        transID = res["TransformationID"]
        records.append(
            [
                transName,
                status,
                str(files_Processed),
                str(files_PercentProcessed),
                str(transID),
                str(prodID),
                str(trans["LastUpdate"]),
                str(trans["InsertedTime"]),
            ]
        )

    printTable(fields, records)

    DIRAC.exit(0)


if __name__ == "__main__":
    main()



""" Test TimeLeft utility

    (Partially) tested here are SGE and LSF, PBS is TO-DO
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pytest
from six.moves import reload_module

from DIRAC import S_OK, gLogger
from DIRAC.Resources.Computing.BatchSystems.TimeLeft.TimeLeft import TimeLeft


gLogger.setLevel("DEBUG")

SGE_OUT = """==============================================================
job_number:                 12345
exec_file:                  job_scripts/12345
submission_time:            Wed Apr 11 09:36:41 2012
owner:                      lhcb049
uid:                        18416
group:                      lhcb
gid:                        155
sge_o_home:                 /home/lhcb049
sge_o_log_name:             lhcb049
sge_o_path:                 /opt/sge/bin/lx24-amd64:/usr/bin:/bin
sge_o_shell:                /bin/sh
sge_o_workdir:              /var/glite/tmp
sge_o_host:                 cccreamceli05
account:                    GRID=EGI SITE=IN2P3-CC TIER=tier1 VO=lhcb ROLEVOMS=&2Flhcb&2FRole=pilot&2FCapability=NULL
merge:                      y
hard resource_list:         os=sl5,s_cpu=1000,s_vmem=5120M,s_fsize=51200M,cvmfs=1,dcache=1
mail_list:                  lhcb049@cccreamceli05.in2p3.fr
notify:                     FALSE
job_name:                   cccreamceli05_crm05_749996134
stdout_path_list:           NONE:NONE:/dev/null
jobshare:                   0
hard_queue_list:            huge
restart:                    n
shell_list:                 NONE:/bin/bash
env_list:                   SITE_NAME=IN2P3-CC,MANPATH=/opt/sge/man:/usr/share/man:/usr/local/man:/usr/local/share/man
script_file:                /tmp/crm05_749996134
project:                    P_lhcb_pilot
usage    1:                 cpu=00:01:00, mem=0.03044 GBs, io=0.19846, vmem=288.609M, maxvmem=288.609M
scheduling info:            (Collecting of scheduler job information is turned off)"""

PBS_OUT = "bla"

LSF_OUT = (
    "JOBID     USER    STAT  QUEUE      FROM_HOST   EXEC_HOST   JOB_NAME   SUBMIT_TIME  PROJ_NAME CPU_USED MEM"
    " SWAP PIDS START_TIME FINISH_TIME\n"
    "12345 user RUN   q1  host1 p01 job1 12/31-20:51:42 default"
    "    00:00:60.00 6267 40713 25469,14249 12/31-20:52:00 -"
)

MJF_OUT = "0"

SLURM_OUT_0 = "12345,86400,24,3600,03:00:00"
SLURM_OUT_1 = "12345,86400,24,3600,4-03:00:00"
SLURM_OUT_2 = "12345,21600,24,900,30:00"
SLURM_OUT_3 = "12345,43200,24,1800,30:00"
SLURM_OUT_4 = ""

HTCONDOR_OUT_0 = "86400 3600"
HTCONDOR_OUT_1 = "undefined 3600"
HTCONDOR_OUT_2 = ""


@pytest.mark.parametrize(
    "batch, requiredVariables, returnValue, expected",
    [
        ("LSF", {}, LSF_OUT, 0.0),
        ("LSF", {"bin": "/usr/bin", "hostNorm": 10.0}, LSF_OUT, 0.0),
        ("MJF", {}, MJF_OUT, 0.0),
        ("SGE", {}, SGE_OUT, 300.0),
        ("SLURM", {}, SLURM_OUT_0, 432000.0),
        ("SLURM", {}, SLURM_OUT_1, 432000.0),
        ("SLURM", {}, SLURM_OUT_2, 108000.0),
        ("SLURM", {}, SLURM_OUT_3, 216000.0),
        ("SLURM", {}, SLURM_OUT_4, 0.0),
        ("HTCondor", {}, HTCONDOR_OUT_0, 18000.0),
        ("HTCondor", {}, HTCONDOR_OUT_1, 0.0),
        ("HTCondor", {}, HTCONDOR_OUT_2, 0.0),
    ],
)
def test_getScaledCPU(mocker, batch, requiredVariables, returnValue, expected):
    """Test getScaledCPU()"""
    mocker.patch("DIRAC.Resources.Computing.BatchSystems.TimeLeft.TimeLeft.runCommand", return_value=S_OK(returnValue))
    tl = TimeLeft()
    res = tl.getScaledCPU()
    assert res == 0

    tl.cpuPower = 5.0

    batchSystemName = "%sResourceUsage" % batch
    batchSystemPath = "DIRAC.Resources.Computing.BatchSystems.TimeLeft.%s" % batchSystemName
    batchPlugin = __import__(batchSystemPath, globals(), locals(), [batchSystemName])  # pylint: disable=unused-variable
    # Need to be reloaded to update the mock within the module, else, it will reuse the one when loaded the first time
    reload_module(batchPlugin)

    batchStr = "batchPlugin.%s()" % (batchSystemName)
    tl.batchPlugin = eval(batchStr)

    # Update attributes of the batch systems to get scaled CPU
    tl.batchPlugin.__dict__.update(requiredVariables)

    res = tl.getScaledCPU()
    assert res == expected


@pytest.mark.parametrize(
    "batch, requiredVariables, returnValue, expected_1, expected_2",
    [
        ("LSF", {"bin": "/usr/bin", "hostNorm": 10.0, "cpuLimit": 1000, "wallClockLimit": 1000}, LSF_OUT, True, 9400.0),
        ("SGE", {}, SGE_OUT, True, 9400.0),
        ("SLURM", {}, SLURM_OUT_0, True, 72000.0),
        ("SLURM", {}, SLURM_OUT_1, True, 3528000.0),
        ("SLURM", {}, SLURM_OUT_2, True, 9000.0),
        ("SLURM", {}, SLURM_OUT_3, True, 0.0),
        ("SLURM", {}, SLURM_OUT_4, False, 0.0),
        ("HTCondor", {}, HTCONDOR_OUT_0, True, 828000),
        ("HTCondor", {}, HTCONDOR_OUT_1, False, 0.0),
        ("HTCondor", {}, HTCONDOR_OUT_2, False, 0.0),
    ],
)
def test_getTimeLeft(mocker, batch, requiredVariables, returnValue, expected_1, expected_2):
    """Test getTimeLeft()"""
    mocker.patch("DIRAC.Resources.Computing.BatchSystems.TimeLeft.TimeLeft.runCommand", return_value=S_OK(returnValue))
    tl = TimeLeft()

    batchSystemName = "%sResourceUsage" % batch
    batchSystemPath = "DIRAC.Resources.Computing.BatchSystems.TimeLeft.%s" % batchSystemName
    batchPlugin = __import__(batchSystemPath, globals(), locals(), [batchSystemName])
    # Need to be reloaded to update the mock within the module, else, it will reuse the one when loaded the first time
    reload_module(batchPlugin)

    batchStr = "batchPlugin.%s()" % (batchSystemName)
    tl.batchPlugin = eval(batchStr)
    tl.cpuPower = 10.0

    # Update attributes of the batch systems to get scaled CPU
    tl.batchPlugin.__dict__.update(requiredVariables)

    res = tl.getTimeLeft()
    assert res["OK"] is expected_1
    if res["OK"]:
        assert res["Value"] == expected_2



""" This class is a wrap around the rrdtool as it is a command line based tool within this class there are
    several methods which take in some parameters required by the corresponding rrd command and executes it.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os.path
import hashlib
from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.PathFinder import getServiceSection
from DIRAC.FrameworkSystem.private.monitoring.ColorGenerator import ColorGenerator
from DIRAC.Core.Utilities import Subprocess, Time
from DIRAC.Core.Utilities.File import mkDir

__RCSID__ = "$Id$"


class RRDManager(object):

    __sizesList = [[200, 50], [400, 100], [600, 150], [800, 200]]
    __logRRDCommands = False

    def __init__(self, rrdLocation, graphLocation):
        """
        Initialize RRDManager
        """
        self.rrdLocation = rrdLocation
        self.graphLocation = graphLocation
        self.log = gLogger.getSubLogger("RRDManager")
        self.rrdExec = gConfig.getValue("%s/RRDExec" % getServiceSection("Framework/Monitoring"), "rrdtool")
        for path in (self.rrdLocation, self.graphLocation):
            mkDir(path)

    def existsRRDFile(self, rrdFile):
        """Checks whether a given rrd file exists or not.

        :type rrdFile: string
        :param rrdFile: name of the rrd file.
        :return: bool
        """
        rrdFilePath = "%s/%s" % (self.rrdLocation, rrdFile)
        return os.path.isfile(rrdFilePath)

    def getGraphLocation(self):
        """
        Sets the location for graph files
        """
        return self.graphLocation

    def __exec(self, cmd, rrdFile=None):
        """
        Executes a system command.

        :type cmd: string
        :param cmd: The cmd command to be executed.
        :type rrdFile: string
        :param rrdFile: name of the rrd file.
        :return: The value dictionary / S_ERROR with a message.
        """
        self.log.debug("RRD command: %s" % cmd)
        retVal = Subprocess.shellCall(0, cmd)
        if self.__logRRDCommands and rrdFile:
            try:
                logFile = "%s.log" % rrdFile
                with open(logFile, "a") as fd:
                    if not retVal["OK"] or retVal["Value"][0]:
                        fd.write("ERROR %s\n" % cmd)
                    else:
                        fd.write("OK    %s\n" % cmd)
            except Exception as e:
                self.log.warn("Cannot write log %s: %s" % (logFile, str(e)))
        if not retVal["OK"]:
            return retVal
        retTuple = retVal["Value"]
        if retTuple[0]:
            return S_ERROR("Failed to execute rrdtool: %s" % (retTuple[2]))
        return retVal

    def getCurrentBucketTime(self, bucketLength):
        """
        Gets current time "bucketized"
        """
        return self.bucketize(Time.toEpoch(), bucketLength)

    def bucketize(self, secs, bucketLength):
        """
        Bucketizes a time (in secs)
        """
        secs = int(secs)
        return secs - secs % bucketLength

    def create(self, type, rrdFile, bucketLength):
        """
        Creates an rrd file.

        :type rrdFile: string
        :param rrdFile: name of the rrd file.
        :type bucketLength: int
        :param bucketLength: The required bucket length.
        :return: The value dictionary / S_ERROR with a message.
        """
        # Understanding this method it basically takes in the activity data which is sent from the client i.e. the gMonitor
        # object and checks if the file is created by the MonitoringCatalog as it creates an .rrd file using the
        # registerActivity method within it.

        rrdFilePath = "%s/%s" % (self.rrdLocation, rrdFile)
        if os.path.isfile(rrdFilePath):
            return S_OK()
        try:
            os.makedirs(os.path.dirname(rrdFilePath))
        except Exception:
            pass
        self.log.info("Creating rrd file %s" % rrdFile)
        cmd = "%s create '%s'" % (self.rrdExec, rrdFilePath)
        # Start GMT(now) - 1h
        cmd += " --start %s" % (self.getCurrentBucketTime(bucketLength) - 86400)
        cmd += " --step %s" % bucketLength
        if type in ("mean"):
            dst = "GAUGE"
            cf = "AVERAGE"
        elif type in ("sum", "acum", "rate"):
            dst = "ABSOLUTE"
            cf = "AVERAGE"
        cmd += " DS:value:%s:%s:U:U" % (dst, bucketLength * 10)
        # 1m res for 1 month
        # cmd += " RRA:%s:0.9:1:43200" % cf
        # 1m red for 1 year
        cmd += " RRA:%s:0.999:1:%s" % (cf, int(31536000 / bucketLength))
        return self.__exec(cmd, rrdFilePath)

    def __getLastUpdateTime(self, rrdFile):
        """
        Gets last update time from an rrd.

        :type rrdFile: string
        :param rrdFile: name of the rrd file.
        :return: S_OK / S_ERROR with a message.
        """
        cmd = "%s last %s" % (self.rrdExec, rrdFile)
        retVal = Subprocess.shellCall(0, cmd)
        if not retVal["OK"]:
            return retVal
        retTuple = retVal["Value"]
        if retTuple[0]:
            return S_ERROR("Failed to fetch last update %s : %s" % (rrdFile, retTuple[2]))
        return S_OK(int(retTuple[1].strip()))

    def __fillWithZeros(self, lastUpdateTime, bucketLength, valuesList):
        filledList = []
        expectedUpdateTime = lastUpdateTime + bucketLength
        for valueTuple in valuesList:
            while expectedUpdateTime < valueTuple[0]:
                filledList.append((expectedUpdateTime, 0))
                expectedUpdateTime += bucketLength
            filledList.append(valueTuple)
            expectedUpdateTime = valueTuple[0] + bucketLength
        return filledList

    def update(self, type, rrdFile, bucketLength, valuesList, lastUpdate=0):
        """
        Updates an rrd file.

        :type rrdFile: string
        :param rrdFile: name of the rrd file.
        :type bucketLength: int
        :param bucketLength: The required bucket length.
        :type valuesList: list
        :param valuesList: a list of values to be updated.
        :type lastUpdate: int
        :param lastUpdate: The timestamp of the last update made to the rrd file.
        :return: S_OK with the updated values list.
        """
        rrdFilePath = "%s/%s" % (self.rrdLocation, rrdFile)
        self.log.info("Updating rrd file", rrdFilePath)
        if lastUpdate == 0:
            retVal = self.__getLastUpdateTime(rrdFilePath)
            if retVal["OK"]:
                lastUpdateTime = retVal["Value"]
                self.log.verbose("Last update time is %s" % lastUpdateTime)
        else:
            lastUpdateTime = lastUpdate
        cmd = "%s update %s" % (self.rrdExec, rrdFilePath)
        # we have to fill with 0 the db to ensure the mean is valid
        valuesList = self.__fillWithZeros(lastUpdateTime, bucketLength, valuesList)
        rrdUpdates = []
        for entry in valuesList:
            rrdUpdates.append("%s:%s" % entry)
        maxRRDArgs = 50
        for i in range(0, len(rrdUpdates), maxRRDArgs):
            finalCmd = "%s %s" % (cmd, " ".join(rrdUpdates[i : i + maxRRDArgs]))
            retVal = self.__exec(finalCmd, rrdFilePath)
            if not retVal["OK"]:
                self.log.warn("Error updating rrd file", "%s rrd: %s" % (rrdFile, retVal["Message"]))
        return S_OK(valuesList[-1][0])

    def __generateName(self, *args, **kwargs):
        """
        Generates a random name
        """
        m = hashlib.md5()
        m.update(str(args).encode())
        m.update(str(kwargs).encode())
        return m.hexdigest()

    def __generateRRDGraphVar(self, entryName, activity, timeSpan, plotWidth):
        """
        Calculates the graph query in rrd lingo for an activity
        """
        rrdFile = activity.getFile()
        rrdType = activity.getType()
        bucketLength = activity.getBucketLength()
        yScaleFactor = self.__getYScalingFactor(timeSpan, bucketLength, plotWidth)
        activity.setBucketScaleFactor(yScaleFactor)
        varStr = "'DEF:ac%sRAW=%s/%s:value:AVERAGE'" % (entryName, self.rrdLocation, rrdFile)
        if rrdType in ("mean", "rate"):
            varStr += " 'CDEF:%s=ac%sRAW,UN,0,ac%sRAW,IF'" % (entryName, entryName, entryName)
        elif rrdType == "sum":
            scale = yScaleFactor * bucketLength
            varStr += " 'CDEF:%s=ac%sRAW,UN,0,ac%sRAW,%s,*,IF'" % (entryName, entryName, entryName, scale)
        elif rrdType == "acum":
            scale = yScaleFactor * bucketLength
            varStr += " 'CDEF:ac%sNOTUN=ac%sRAW,UN,0,ac%sRAW,%s,*,IF'" % (entryName, entryName, entryName, scale)
            varStr += " 'CDEF:%s=PREV,UN,ac%sNOTUN,PREV,ac%sNOTUN,+,IF'" % (entryName, entryName, entryName)
        return varStr

    def __graphTimeComment(self, fromEpoch, toEpoch):
        comStr = " 'COMMENT:Generated on %s UTC'" % Time.toString().replace(":", r"\:").split(".")[0]
        comStr += " 'COMMENT:%s'" % ("From %s to %s" % (Time.fromEpoch(fromEpoch), Time.fromEpoch(toEpoch))).replace(
            ":", r"\:"
        )
        return comStr

    def __getYScalingFactor(self, timeSpan, bucketLength, plotWidth):
        expectedTimeSpan = plotWidth * bucketLength
        if timeSpan < expectedTimeSpan:
            return 1
        else:
            return float(timeSpan) / expectedTimeSpan

    def groupPlot(self, fromSecs, toSecs, activitiesList, stackActivities, size, graphFilename=""):
        """
        Generates a group plot.

        :type fromSecs: int
        :param fromSecs: A value in seconds from where to start.
        :type toSecs: int
        :param toSecs: A value in seconds for where to end.
        :type activitiesList: list
        :param activitiesList: A list of activities.
        :type stackActivities: list
        :param stackActivities: A list of stacked activities.
        :type size: int
        :param size: There is a matrix defined for size so here only one of these values go [0, 1, 2, 3].
        :type graphFilename: string
        :param graphFilename: A name for the graph file.
        :return: S_OK with the graph filename / The error message.
        """
        plotTimeSpan = toSecs - fromSecs
        if not graphFilename:
            graphFilename = "%s.png" % self.__generateName(fromSecs, toSecs, activitiesList, stackActivities)
        rrdCmd = "%s graph %s/%s" % (self.rrdExec, self.graphLocation, graphFilename)
        rrdCmd += " -s %s" % fromSecs
        rrdCmd += " -e %s" % toSecs
        rrdCmd += " -w %s" % self.__sizesList[size][0]
        rrdCmd += " -h %s" % self.__sizesList[size][1]
        rrdCmd += " --title '%s'" % activitiesList[0].getGroupLabel()
        colorGen = ColorGenerator()
        activitiesList.sort()
        for idActivity in range(len(activitiesList)):
            activity = activitiesList[idActivity]
            rrdCmd += " %s" % self.__generateRRDGraphVar(idActivity, activity, plotTimeSpan, self.__sizesList[size][0])
            if stackActivities:
                rrdCmd += " 'AREA:%s#%s:%s:STACK'" % (
                    idActivity,
                    colorGen.getHexColor(),
                    activity.getLabel().replace(":", r"\:"),
                )
            else:
                rrdCmd += " 'LINE1:%s#%s:%s'" % (
                    idActivity,
                    colorGen.getHexColor(),
                    activity.getLabel().replace(":", r"\:"),
                )
        rrdCmd += self.__graphTimeComment(fromSecs, toSecs)
        retVal = self.__exec(rrdCmd)
        if not retVal["OK"]:
            return retVal
        return S_OK(graphFilename)

    def plot(self, fromSecs, toSecs, activity, stackActivities, size, graphFilename=""):
        """
        Generates a non grouped plot.

        :type fromSecs: int
        :param fromSecs: A value in seconds from where to start.
        :type toSecs: int
        :param toSecs: A value in seconds for where to end.
        :type activitiesList: list
        :param activitiesList: A list of activities.
        :type stackActivities: list
        :param stackActivities: A list of stacked activities.
        :type size: int
        :param size: There is a matrix defined for size so here only one of these values go [0, 1, 2, 3].
        :type graphFilename: string
        :param graphFilename: A name for the graph file.
        :return: S_OK with the graph filename / The error message.
        """
        plotTimeSpan = toSecs - fromSecs
        if not graphFilename:
            graphFilename = "%s.png" % self.__generateName(fromSecs, toSecs, activity, stackActivities)
        graphVar = self.__generateRRDGraphVar(0, activity, plotTimeSpan, self.__sizesList[size][0])
        rrdCmd = "%s graph %s/%s" % (self.rrdExec, self.graphLocation, graphFilename)
        rrdCmd += " -s %s" % fromSecs
        rrdCmd += " -e %s" % toSecs
        rrdCmd += " -w %s" % self.__sizesList[size][0]
        rrdCmd += " -h %s" % self.__sizesList[size][1]
        rrdCmd += " --title '%s'" % activity.getLabel()
        rrdCmd += " --vertical-label '%s'" % activity.getUnit()
        rrdCmd += " %s" % graphVar
        if stackActivities:
            rrdCmd += " 'AREA:0#0000FF::STACK'"
        else:
            rrdCmd += " 'LINE1:0#0000FF'"
        rrdCmd += self.__graphTimeComment(fromSecs, toSecs)
        retVal = self.__exec(rrdCmd)
        if not retVal["OK"]:
            return retVal
        return S_OK(graphFilename)

    def deleteRRD(self, rrdFile):
        """This method is used to delete an rrd file.

        :type rrdFile: string
        :param rrdFile: name of the rrd file.
        """
        try:
            os.unlink("%s/%s" % (self.rrdLocation, rrdFile))
        except Exception as e:
            self.log.error("Could not delete rrd file", "%s: %s" % (rrdFile, str(e)))



#!/usr/bin/env python

""" Performs a DIPS ping on a given target and exit with the return code.
It uses the local host certificate
The target is specified as ""<port>/System/Service"
The script does not print anything, and just exists with 0 in case of success,
or 1 in case of error """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


import sys
import os
import time
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    with open(os.devnull, "w") as redirectStdout, open(os.devnull, "w") as redirectStderr:
        from DIRAC import gLogger
        from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData

        gConfigurationData.setOptionInCFG("/DIRAC/Security/UseServerCertificate", "true")
        gLogger.setLevel("FATAL")
        from DIRAC.Core.DISET.RPCClient import RPCClient

        rpc = RPCClient("dips://localhost:%s" % sys.argv[1])
        res = rpc.ping()
        time.sleep(0.1)
        sys.exit(0 if res["OK"] else 1)


if __name__ == "__main__":
    main()



########################################################################
# File: RequestExecutingAgent.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2013/03/12 15:36:47
########################################################################

"""Agent processing the requests

 .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

See also the information about the :ref:`requestManagementSystem`.

The following options can be set for the RequestExecutingAgent. The configuration also includes the
``OperationHandlers`` available in DIRAC.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN RequestExecutingAgent
  :end-before: ##END
  :dedent: 2
  :caption: RequestExecutingAgent options

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

# #
# @file RequestExecutingAgent.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2013/03/12 15:36:56
# @brief Definition of RequestExecutingAgent class.

# # imports
import sys
import time
import errno

# # from DIRAC
from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.Core.Utilities.ProcessPool import ProcessPool
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
from DIRAC.RequestManagementSystem.private.RequestTask import RequestTask

from DIRAC.MonitoringSystem.Client.MonitoringReporter import MonitoringReporter
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler
from DIRAC.Core.Utilities import Time, Network

from DIRAC.Core.Utilities.DErrno import cmpError

# # agent name
AGENT_NAME = "RequestManagement/RequestExecutingAgent"


class AgentConfigError(Exception):
    """misconfiguration error"""

    def __init__(self, msg):
        """ctor

        :param str msg: error string
        """
        Exception.__init__(self)
        self.msg = msg

    def __str__(self):
        """str op"""
        return self.msg


########################################################################


class RequestExecutingAgent(AgentModule):
    """
    .. class:: RequestExecutingAgent

    request processing agent using ProcessPool, Operation handlers and RequestTask
    """

    # # process pool
    __processPool = None
    # # request cache
    __requestCache = {}
    # # requests/cycle
    __requestsPerCycle = 100
    # # minimal nb of subprocess running
    __minProcess = 20
    # # maximal nb of subprocess executed same time
    __maxProcess = 20
    # # ProcessPool queue size
    __queueSize = 20
    # # file timeout
    __fileTimeout = 300
    # # operation timeout
    __operationTimeout = 300
    # # ProcessPool finalization timeout
    __poolTimeout = 900
    # # ProcessPool sleep time
    __poolSleep = 5
    # # placeholder for RequestClient instance
    __requestClient = None
    # # Size of the bulk if use of getRequests. If 0, use getRequest
    __bulkRequest = 0
    # # Send the monitoring data to ES rather than the Framework/Monitoring
    __rmsMonitoring = False

    def __init__(self, *args, **kwargs):
        """c'tor"""
        # # call base class ctor
        AgentModule.__init__(self, *args, **kwargs)
        # # ProcessPool related stuff
        self.__requestsPerCycle = self.am_getOption("RequestsPerCycle", self.__requestsPerCycle)
        self.log.info("Requests/cycle = %d" % self.__requestsPerCycle)
        self.__minProcess = self.am_getOption("MinProcess", self.__minProcess)
        self.log.info("ProcessPool min process = %d" % self.__minProcess)
        self.__maxProcess = self.am_getOption("MaxProcess", self.__maxProcess)
        self.log.info("ProcessPool max process = %d" % self.__maxProcess)
        self.__queueSize = self.am_getOption("ProcessPoolQueueSize", self.__queueSize)
        self.log.info("ProcessPool queue size = %d" % self.__queueSize)
        self.__poolTimeout = int(self.am_getOption("ProcessPoolTimeout", self.__poolTimeout))
        self.log.info("ProcessPool timeout = %d seconds" % self.__poolTimeout)
        self.__poolSleep = int(self.am_getOption("ProcessPoolSleep", self.__poolSleep))
        self.log.info("ProcessPool sleep time = %d seconds" % self.__poolSleep)
        self.__bulkRequest = self.am_getOption("BulkRequest", self.__bulkRequest)
        self.log.info("Bulk request size = %d" % self.__bulkRequest)
        self.__rmsMonitoring = self.am_getOption("EnableRMSMonitoring", self.__rmsMonitoring)
        self.log.info("Enable ES RMS Monitoring = %s" % self.__rmsMonitoring)

        # # keep config path and agent name
        self.agentName = self.am_getModuleParam("fullName")
        self.__configPath = PathFinder.getAgentSection(self.agentName)

        # # operation handlers over here
        opHandlersPath = "%s/%s" % (self.__configPath, "OperationHandlers")
        opHandlers = gConfig.getSections(opHandlersPath)
        if not opHandlers["OK"]:
            self.log.error(opHandlers["Message"])
            raise AgentConfigError("OperationHandlers section not found in CS under %s" % self.__configPath)
        opHandlers = opHandlers["Value"]

        self.timeOuts = dict()

        # # handlers dict
        self.handlersDict = dict()
        for opHandler in opHandlers:
            opHandlerPath = "%s/%s/Location" % (opHandlersPath, opHandler)
            opLocation = gConfig.getValue(opHandlerPath, "")
            if not opLocation:
                self.log.error("%s not set for %s operation handler" % (opHandlerPath, opHandler))
                continue
            self.timeOuts[opHandler] = {"PerFile": self.__fileTimeout, "PerOperation": self.__operationTimeout}

            opTimeout = gConfig.getValue("%s/%s/TimeOut" % (opHandlersPath, opHandler), 0)
            if opTimeout:
                self.timeOuts[opHandler]["PerOperation"] = opTimeout
            fileTimeout = gConfig.getValue("%s/%s/TimeOutPerFile" % (opHandlersPath, opHandler), 0)
            if fileTimeout:
                self.timeOuts[opHandler]["PerFile"] = fileTimeout

            self.handlersDict[opHandler] = opLocation

        self.log.info("Operation handlers:")
        for item in enumerate(self.handlersDict.items()):
            opHandler = item[1][0]
            self.log.info(
                "[%s] %s: %s (timeout: %d s + %d s per file)"
                % (
                    item[0],
                    item[1][0],
                    item[1][1],
                    self.timeOuts[opHandler]["PerOperation"],
                    self.timeOuts[opHandler]["PerFile"],
                )
            )

        if self.__rmsMonitoring:
            self.rmsMonitoringReporter = MonitoringReporter(monitoringType="RMSMonitoring")
            gThreadScheduler.addPeriodicTask(100, self.__rmsMonitoringReporting)
        else:
            # # common monitor activity
            gMonitor.registerActivity("Iteration", "Agent Loops", "RequestExecutingAgent", "Loops/min", gMonitor.OP_SUM)
            gMonitor.registerActivity(
                "Processed", "Request Processed", "RequestExecutingAgent", "Requests/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "Done", "Request Completed", "RequestExecutingAgent", "Requests/min", gMonitor.OP_SUM
            )

        # # create request dict
        self.__requestCache = dict()

        # ?? Probably should be removed
        self.FTSMode = self.am_getOption("FTSMode", False)

    def processPool(self):
        """facade for ProcessPool"""
        if not self.__processPool:
            minProcess = max(1, self.__minProcess)
            maxProcess = max(self.__minProcess, self.__maxProcess)
            queueSize = abs(self.__queueSize)
            self.log.info(
                "REA ProcessPool configuration",
                "minProcess = %d maxProcess = %d queueSize = %d" % (minProcess, maxProcess, queueSize),
            )
            self.__processPool = ProcessPool(
                minProcess,
                maxProcess,
                queueSize,
                poolCallback=self.resultCallback,
                poolExceptionCallback=self.exceptionCallback,
            )
            self.__processPool.daemonize()
        return self.__processPool

    def requestClient(self):
        """RequestClient getter"""
        if not self.__requestClient:
            self.__requestClient = ReqClient()
        return self.__requestClient

    def cacheRequest(self, request):
        """put request into requestCache

        :param ~Request.Request request: Request instance
        """
        maxProcess = max(self.__minProcess, self.__maxProcess)
        if len(self.__requestCache) > maxProcess + 50:
            # For the time being we just print a warning... If the ProcessPool is working well, this is not needed
            # We don't know how much is acceptable as it depends on many factors
            self.log.warn("Too many requests in cache", ": %d" % len(self.__requestCache))
        #      return S_ERROR( "Too many requests in cache" )
        if request.RequestID in self.__requestCache:
            # We don't call  putRequest as we have got back the request that is still being executed. Better keep it
            # The main reason for this is that it lasted longer than the kick time of CleanReqAgent
            self.log.warn(
                "Duplicate request, keep it but don't execute", ": %d/%s" % (request.RequestID, request.RequestName)
            )
            return S_ERROR(errno.EALREADY, "Request already in cache")
        self.__requestCache[request.RequestID] = request
        return S_OK()

    def putRequest(self, requestID, taskResult=None):
        """put back :requestID: to RequestClient

        :param str requestID: request's id
        """
        if requestID in self.__requestCache:
            request = self.__requestCache.pop(requestID)
            if taskResult:
                if taskResult["OK"]:
                    request = taskResult["Value"]
                    # The RequestTask is putting back the Done tasks, no need to redo it
                    if request.Status == "Done":
                        return S_OK()
                # In case of timeout, we need to increment ourselves all the attempts
                elif cmpError(taskResult, errno.ETIME):
                    waitingOp = request.getWaiting()
                    for rmsFile in waitingOp.get("Value", []):
                        rmsFile.Attempt += 1

            reset = self.requestClient().putRequest(request, useFailoverProxy=False, retryMainService=2)
            if not reset["OK"]:
                return S_ERROR("putRequest: unable to reset request %s: %s" % (requestID, reset["Message"]))
        else:
            return S_ERROR("Not in cache")
        return S_OK()

    def putAllRequests(self):
        """put back all requests without callback called into requestClient

        :param self: self reference
        """
        self.log.info("putAllRequests: will put back requests", "%s" % len(self.__requestCache))
        for requestID in self.__requestCache.keys():
            reset = self.putRequest(requestID)
            if not reset["OK"]:
                self.log.error("Failed to put request", reset["Message"])
            else:
                self.log.debug("putAllRequests: request %s has been put back with its initial state" % requestID)
        return S_OK()

    def initialize(self):
        """initialize agent"""
        return S_OK()

    def execute(self):
        """read requests from RequestClient and enqueue them into ProcessPool"""
        if not self.__rmsMonitoring:
            gMonitor.addMark("Iteration", 1)
        # # requests (and so tasks) counter
        taskCounter = 0
        while taskCounter < self.__requestsPerCycle:
            self.log.debug("execute: executing %d request in this cycle" % taskCounter)

            requestsToExecute = []

            if not self.__bulkRequest:
                self.log.info("execute: ask for a single request")
                getRequest = self.requestClient().getRequest()
                if not getRequest["OK"]:
                    self.log.error("execute:", "%s" % getRequest["Message"])
                    break
                if not getRequest["Value"]:
                    self.log.info("execute: no more 'Waiting' requests to process")
                    break
                requestsToExecute = [getRequest["Value"]]
            else:
                numberOfRequest = min(self.__bulkRequest, self.__requestsPerCycle - taskCounter)
                self.log.info("execute: ask for requests", "%s" % numberOfRequest)
                getRequests = self.requestClient().getBulkRequests(numberOfRequest)
                if not getRequests["OK"]:
                    self.log.error("execute:", "%s" % getRequests["Message"])
                    break
                if not getRequests["Value"]:
                    self.log.info("execute: no more 'Waiting' requests to process")
                    break
                for rId in getRequests["Value"]["Failed"]:
                    self.log.error("execute:", "%s" % getRequests["Value"]["Failed"][rId])

                requestsToExecute = list(getRequests["Value"]["Successful"].values())

            self.log.info("execute: will execute requests ", "%s" % len(requestsToExecute))

            for request in requestsToExecute:
                # # set task id
                taskID = request.RequestID

                self.log.info(
                    "processPool status",
                    "tasks idle = %s working = %s"
                    % (self.processPool().getNumIdleProcesses(), self.processPool().getNumWorkingProcesses()),
                )

                looping = 0
                while True:
                    if not self.processPool().getFreeSlots():
                        if not looping:
                            self.log.info(
                                "No free slots available in processPool",
                                "will wait %d seconds to proceed" % self.__poolSleep,
                            )
                        time.sleep(self.__poolSleep)
                        looping += 1
                    else:
                        if looping:
                            self.log.info("Free slot found", "after %d seconds" % looping * self.__poolSleep)
                        looping = 0
                        # # save current request in cache
                        res = self.cacheRequest(request)
                        if not res["OK"]:
                            if cmpError(res, errno.EALREADY):
                                # The request is already in the cache, skip it. break out of the while loop to get next request
                                break
                            # There are too many requests in the cache, commit suicide
                            self.log.error(
                                "Too many requests in cache",
                                "(%d requests): put back all requests and exit cycle. Error %s"
                                % (len(self.__requestCache), res["Message"]),
                            )
                            self.putAllRequests()
                            return res
                        # # serialize to JSON
                        result = request.toJSON()
                        if not result["OK"]:
                            continue
                        requestJSON = result["Value"]
                        self.log.info("spawning task for request", "'%s/%s'" % (request.RequestID, request.RequestName))
                        timeOut = self.getTimeout(request)
                        enqueue = self.processPool().createAndQueueTask(
                            RequestTask,
                            kwargs={
                                "requestJSON": requestJSON,
                                "handlersDict": self.handlersDict,
                                "csPath": self.__configPath,
                                "agentName": self.agentName,
                                "rmsMonitoring": self.__rmsMonitoring,
                            },
                            taskID=taskID,
                            blocking=True,
                            usePoolCallbacks=True,
                            timeOut=timeOut,
                        )
                        if not enqueue["OK"]:
                            self.log.error("Could not enqueue task", enqueue["Message"])
                        else:
                            self.log.debug("successfully enqueued task", "'%s'" % taskID)
                            # # update monitor
                            if self.__rmsMonitoring:
                                self.rmsMonitoringReporter.addRecord(
                                    {
                                        "timestamp": int(Time.toEpoch()),
                                        "host": Network.getFQDN(),
                                        "objectType": "Request",
                                        "status": "Attempted",
                                        "objectID": request.RequestID,
                                        "nbObject": 1,
                                    }
                                )
                            else:
                                gMonitor.addMark("Processed", 1)

                            # # update request counter
                            taskCounter += 1
                            # # task created, a little time kick to proceed
                            time.sleep(0.1)
                            break

        self.log.info("Flushing callbacks", "(%d requests still in cache)" % len(self.__requestCache))
        processed = self.processPool().processResults()
        # This happens when the result queue is screwed up.
        # Returning S_ERROR proved not to be sufficient,
        # and when in this situation, there is nothing we can do.
        # So we just exit. runit will restart from scratch.
        if processed < 0:
            self.log.fatal("Results queue is screwed up")
            sys.exit(1)
        # # clean return
        return S_OK()

    def getTimeout(self, request):
        """get timeout for request"""
        timeout = 0
        for op in request:
            if op.Status not in ("Waiting", "Scheduled", "Queued"):
                continue
            if op.Type not in self.timeOuts:
                timeout += self.__operationTimeout
            else:
                perOp = self.timeOuts[op.Type].get("PerOperation", self.__operationTimeout)
                perFiles = self.timeOuts[op.Type].get("PerFile", self.__fileTimeout) * len(op)
                timeout += perOp + perFiles
        self.log.info(
            "estimated timeOut for request", "(%s/%s) is %s" % (request.RequestID, request.RequestName, timeout)
        )
        return timeout

    def finalize(self):
        """agent finalization"""
        if self.__processPool:
            self.processPool().finalize(timeout=self.__poolTimeout)
        self.putAllRequests()
        return S_OK()

    def resultCallback(self, taskID, taskResult):
        """definition of request callback function

        :param str taskID: Request.RequestID
        :param dict taskResult: task result S_OK(Request)/S_ERROR(Message)
        """
        # # clean cache
        res = self.putRequest(taskID, taskResult)
        self.log.info(
            "callback:",
            "%s result is %s(%s), put %s(%s)"
            % (
                taskID,
                "S_OK" if taskResult["OK"] else "S_ERROR",
                taskResult["Value"].Status if taskResult["OK"] else taskResult["Message"],
                "S_OK" if res["OK"] else "S_ERROR",
                "" if res["OK"] else res["Message"],
            ),
        )

    def exceptionCallback(self, taskID, taskException):
        """definition of exception callback function

        :param str taskID: Request.RequestID
        :param Exception taskException: Exception instance
        """
        self.log.error("exceptionCallback:", "%s was hit by exception %s" % (taskID, taskException))
        self.putRequest(taskID)

    def __rmsMonitoringReporting(self):
        """This method is called by the ThreadScheduler as a periodic task in order to commit the collected data which
        is done by the MonitoringReporter and is send to the 'RMSMonitoring' type.
        :return: True / False
        """
        result = self.rmsMonitoringReporter.commit()
        return result["OK"]
